"""

    ConcFactorsEntry.py

    Copyright (c) 2020-2023, SAXS Team, KEK-PF

"""
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting, get_beamline_name
from molass_legacy.KekLib.OurTkinter import Tk

class ConcFactorsEntry:
    def __init__(self, frame, grid_row):
        self.path_length = Tk.StringVar()
        self.extinction = Tk.StringVar()
        self.beamline_name = Tk.StringVar()
        self.reset_entries()

        self.widgets = []

        # UV Device No
        device_label = Tk.Label(frame, text='Beamline: ')
        device_label.grid(row=grid_row, column=0, sticky=Tk.E)
        self.widgets.append(device_label)

        device_no_label = Tk.Entry(frame, textvariable=self.beamline_name, width=20)
        device_no_label.grid(row=grid_row, column=1, sticky=Tk.W )
        self.widgets.append(device_no_label)

        # Conc. Factors
        row = grid_row+1
        conc_label = Tk.Label( frame, text='Conc. Factors: ' )
        conc_label.grid( row=row, column=0, sticky=Tk.E )
        self.widgets.append(conc_label)

        conc_factor_frame = Tk.Frame( frame )
        conc_factor_frame.grid( row=row, column=1, columnspan=2, sticky=Tk.W )

        conc_factor_label1 = Tk.Label( conc_factor_frame, text='Absorbance(λ=%g) × Path Length Factor(' % ( get_setting( 'absorbance_picking' ) ) )
        conc_factor_label1.grid( row=0, column=0 )
        self.widgets.append(conc_factor_label1)

        path_length_entry = Tk.Entry( conc_factor_frame, textvariable=self.path_length, width=6, justify=Tk.CENTER )
        path_length_entry.grid( row=0, column=1 )
        self.widgets.append(path_length_entry)

        conc_factor_label2 = Tk.Label( conc_factor_frame, text=') ÷ Extinction Coefficient(' )
        conc_factor_label2.grid( row=0, column=2 )
        self.widgets.append(conc_factor_label2)

        extinction_entry = Tk.Entry( conc_factor_frame, textvariable=self.extinction, width=7, justify=Tk.CENTER )
        extinction_entry.grid( row=0, column=3 )
        self.widgets.append(extinction_entry)

        conc_factor_label3 = Tk.Label( conc_factor_frame, text=')' )
        conc_factor_label3.grid( row=0, column=4 )
        self.widgets.append(conc_factor_label3)

    def config(self, **kwargs):
        for w in self.widgets:
            w.config(**kwargs)

    def compute_conc_factor(self):
        try:
            path_length = float(self.path_length.get())
            extinction = float(self.extinction.get())
            conc_factor = path_length/extinction
        except:
            raise ValueError

        set_setting('path_length', path_length)
        set_setting('extinction', extinction)
        set_setting('conc_factor', conc_factor)
        return conc_factor

    def update_conc_factors(self, sd, mapper):
        conc_factor = self.compute_conc_factor()
        sd.set_mc_vector(mapper, conc_factor)

    def update_path_length(self):
        path_length = get_setting( 'path_length' )
        self.path_length.set( '%.3f' % path_length )
        self.beamline_name.set(get_beamline_name())

    def reset_entries(self):
        path_length = get_setting( 'path_length' )
        self.path_length.set( '%.3f' % path_length )
        extinction = get_setting( 'extinction' )
        self.extinction.set( '%.3f' % extinction )
        self.beamline_name.set(get_beamline_name())

    def apply_entries(self):
        set_setting('beamline_name', self.beamline_name.get())      # used in SummaryBook



"""

    Peaks.ElutionModels.py

    Copyright (c) 2017-2023, SAXS Team, KEK-PF

"""
import numpy as np
from scipy.special import erfc, erfcx
from scipy.optimize import fsolve

def gaussian(x, h, m, s):
    return h * np.exp(-((x-m)/s)**2)

def egh_impl(x, H, tR, sigma, tau):
    x_  = x - tR
    s2  = 2 * sigma**2
    z   = s2 + tau*x_
    z_neg   = z <= 0
    z_pos   = z > 0

    zero_part = np.zeros( len(x) )[z_neg]
    posi_part = H * np.exp( -x_[z_pos]**2/( s2 + tau*x_[z_pos] ) )

    if tau > 0:
        parts = [ zero_part, posi_part ]
    else:
        parts = [ posi_part, zero_part ]

    return np.hstack( parts )

def egh(x, H, tR, sigma, tau):
    if np.isscalar(x):
        x = np.array([x])
        return egh_impl(x, H, tR, sigma, tau)[0]
    else:
        return egh_impl(x, H, tR, sigma, tau)

"""
Kevin Lan, James W. Jorgenson, Journal of Chromatography A, 915 (2001) 1?13
"""
A0i = np.array([4, -6.293724, 9.232834, -11.342910, 9.123978, -4.173753, 0.827797])
e0 = np.poly1d(A0i[::-1])
"""
e0(th) = a0 + a1*th + a2*th**2 + ... + am*th**m
"""
A1i = np.array([0.75, 0.033807, -0.301080, 1.200371, -1.813317, 1.279318, -0.326582])
e1 = np.poly1d(A1i[::-1])

A2i = np.array([1, -0.982254, 0.568593, 0.512587, -1.184361, 0.939222, -0.240814])
e2 = np.poly1d(A2i[::-1])

# A3i = np.array([0.5, -0.664611, 0.706192, -0.293500, -0.083980, 0.200306, -0.064264])
A3i = np.array([0.49106035, -0.45648717, -0.93606276, 5.25394258, -9.16354757, 7.36039624, -2.25720537])
e3 = np.poly1d(A3i[::-1])

SQRT_PI_8 = np.sqrt(np.pi/8)
def egh_pdf(x, tR=0, sigma=1.0, tau=0.0, scale=1):
    tau_ = abs(tau)
    th = np.arctan2(tau_, sigma)
    H_ = (sigma*SQRT_PI_8 + tau_)*e0(th)
    return egh(x, scale, tR, sigma, tau)/H_

def compute_moments(x, y):
    W = np.sum(y)
    M1 = np.sum(x*y)/W              # raw moment
    M2 = np.sum(y*(x-M1)**2)/W      # central moment
    M3 = np.sum(y*(x-M1)**3)/W      # central moment
    return M1, M2, M3

def compute_moments_from_egh_params(tR, sigma, tau):
    tau_ = abs(tau)
    theta = np.arctan2(tau_, sigma)
    M1 = tR + tau*e1(theta)
    M2 = (sigma**2 + sigma*tau_ + tau**2)*e2(theta)
    M3 = tau*(3*sigma**2 + 4*sigma*tau_ + 4*tau**2)*e3(theta)
    return M1, M2, M3

def compute_egh_params(init_params, moments):
    M1, M2, M3 = moments

    def equations(p):
        tR, sigma, tau = p
        tau_ = abs(tau)         # better for 2019118_3
        # tau_ = tau            # seems better for some cases such as 20181203, although abs(tau) is mentioned in the paper.
        th = np.arctan2(tau_, sigma)
        return (
                    tR + tau*e1(th) - M1,
                    (sigma**2 + sigma*tau_ + tau**2)*e2(th) - M2,
                    tau*(3*sigma**2 + 4*sigma*tau_ + 4*tau**2)*e3(th) - M3,
               )

    x, infodict, ier, mesg = fsolve(equations, init_params, full_output=True)
    if ier in [1, 4, 5]:
        """
        5 : The iteration is not making good progress, as measured by the
            improvement from the last ten iterations.
        4 : The iteration is not making good progress, as measured by the
            improvement from the last five Jacobian evaluations.
        """
        tR, sigma, tau = x
    else:
        assert False
    # print('init_params=', init_params)
    # print('new_params=', (h, tR, sigma, tau))
    return tR, sigma, tau

"""
    This implemetation is based on the following wikipedia page
        https://en.wikipedia.org/wiki/Exponentially_modified_Gaussian_distribution
"""
VERY_SMALL_VALUE    = 1e-8
SQRT2       = 1/np.sqrt(2)
SQRTPI2     = np.sqrt(np.pi/2)
Z_LIMIT     = 6.71e7
LARGE_VALUE = 1e10

def emg_orig(x, h=1, mu=0, sigma=1.0, tau=0):

    if abs(tau) > VERY_SMALL_VALUE:
        z   = SQRT2 * ( sigma/tau - (x - mu)/sigma )
        z_neg   = z < 0
        z_mid   = np.logical_and( z >= 0, z <= Z_LIMIT )
        z_lrg   = z > Z_LIMIT

        ret_neg = h * sigma / tau * SQRTPI2 * np.exp( 0.5*(sigma/tau)**2 - (x[z_neg] - mu)/tau ) * erfc( z[z_neg] )
        ret_mid = h * np.exp( -0.5 * ( (x[z_mid] - mu)/sigma )**2 ) * sigma/tau * SQRTPI2 * erfcx( z[z_mid] )
        x_mu    = x[z_lrg] - mu
        ret_lrg = h * np.exp( -0.5 * ( x_mu/sigma )**2 ) / ( 1 + x_mu*tau/(sigma**2) )
        ret     = np.hstack( [ ret_lrg, ret_mid, ret_neg ] )
    else:
        x_mu    = x - mu
        ret     = h * np.exp( -0.5 * ( x_mu/sigma )**2 ) / ( 1 + x_mu*tau/(sigma**2) )

    return ret

def emg(x, h, mu, sigma, tau):
    # task: not sure if this works for a sclar x
    if tau > 0:
        ret = emg_orig( x, h, mu, sigma, tau )
    else:
        """
            reverse transformation is required
            because emg_orig assumes the ascending order of x
        """
        ret = emg_orig( (mu - x[::-1]), h, 0, sigma, -tau )[::-1]

    if False:
        print( 'tau=', tau )
        fig = dplt.figure()
        ax  = fig.add_subplot(111)
        ax.set_title( "emg" )
        ax.plot( x, ret )
        dplt.show()

    return ret



"""

    Models.EGH.py

    Copyright (c) 2017-2025, SAXS Team, KEK-PF

"""
import numpy as np
from bisect import bisect_right
from scipy.interpolate import UnivariateSpline
from scipy.optimize import minimize
from molass_legacy.KekLib.Affine import Affine
from molass_legacy.Peaks.ElutionModels import egh, VERY_SMALL_VALUE, compute_moments, compute_egh_params
from molass_legacy._MOLASS.SerialSettings import get_setting
from .ElutionModelUtils import x_from_height_ratio_impl
from .ModelUtils import _get_model_name

DEVEL = False
if DEVEL:
    from importlib import reload
    import Models.Tentative
    reload(Models.Tentative)
from .Tentative import Model

# task: these limits should be eventually superseded by TAU_BOUND_RATIO
EGH_TAU_LOWER_LIMIT = -50
EGH_TAU_UPPER_LIMIT = 50

"""
    This implemetation is based on the following paper
        A hybrid of exponential and gaussian functions as a simple model
        of asymmetric chromatographic peaks
        http://acadine.physics.jmu.edu/group/technical_notes/GC_peak_fitting/X_lan_Jorgenson.pdf
"""

def egh_x_from_height_ratio(alpha, tR, sigma, tau):
    alpha_ = np.log(1/alpha)
    v = -tau*alpha_/2
    w = np.sqrt((8*sigma**2 + tau**2*alpha_)*alpha_)/2
    return tR - (v+w), tR - (v-w)

def egh_params( x, params ):
    H   = params['H']
    tR  = params['tR']
    sigma  = params['sigma']
    tau  = params['tau']
    return egh( x, H, tR, sigma, tau )

ALPHA       = 0.1
LN_ALPHA    = np.log(ALPHA)

class EGH(Model):
    def __init__(self, **kwargs):
        Model.__init__(self, egh, **kwargs)

    def guess(self, y, x=None, negative=False, **kwargs):

        n   = np.argmax( y )
        H   = y[n]
        yf  = H * ALPHA

        iL  = bisect_right( y[0:n], yf )
        A   = x[n] - x[iL]
        iR   = bisect_right( list(reversed(y[n:])), yf )
        B   = x[len(x)-1-iR] - x[n]

        if False:
            print( 'iL=', iL, 'A=', A )
            print( 'iR=', iR, 'B=', B )
            print( 'n=', n, 'x[n]=', x[n] )

        tR  = x[n]
        sigma = np.sqrt( -B*A/(2*LN_ALPHA) )
        tau = -(B - A)/LN_ALPHA

        return np.array([H, tR, sigma, tau])

    def get_name( self ):
        return _get_model_name( self.name )

    def x_from_height_ratio(self, ecurve, ratio, params):
        # mu, sigma, tau
        return x_from_height_ratio_impl(egh, ecurve, ratio, *params[1:])

    def get_peaktop_xy(self, x, params):
        mu = params[1]
        py = self.func(mu, *params)
        return mu, py

    def get_params_string(self, params):
        return 'h=%g, mu=%g, sigma=%g, tau=%g' % tuple(params)

    def guess_a_peak_with_prop(self, x, y, prop):
        """
        This function is used to guess a modeled peak with a given area ratio.
        """
        H, tR, sigma, tau = self.guess(y, x=x)
        total_area = np.sum(y)
        def objective(p):
            sigma_, tau_  = p
            cy = egh(x, H, tR, sigma_, tau_)
            area = np.sum(cy)
            return abs(area - total_area*prop)
        
        ret = minimize(objective, [sigma, tau], method='Nelder-Mead')
        sigma, tau = ret.x
        return np.array([H, tR, sigma, tau])

def egha_impl(x, H, tR, sigma, tau, a, raise_=False):
    y   = egh( x, H, tR, sigma, tau )

    if abs(a) < VERY_SMALL_VALUE/max(VERY_SMALL_VALUE, H):
        return y

    xf  = tR - 5*sigma
    xt  = tR + 5*sigma

    src_points  = [ (xf, 0), (tR, H), (xt, 0) ]
    tgt_points  = [ (xf, 0), (tR + a, H ), (xt, 0) ]

    try:
        affine = Affine( src_points, tgt_points, raise_=raise_ )
        x_, y_ = affine.transform( x, y )

        if len(x_) > 1:
            spline = UnivariateSpline( x_, y_, s=0, ext=3 )
            return spline( x )
        else:
            return y_
    except:
        return np.ones(len(x)) * np.nan

def egha(x, H, tR, sigma, tau, a, raise_=False):
    if np.isscalar(x):
        x = np.array([x])
        return egha_impl(x, H, tR, sigma, tau, a, raise_=raise_)[0]
    else:
        return egha_impl(x, H, tR, sigma, tau, a, raise_=raise_)

def egh_no_affine(x, H=1, tR=0, sigma=1.0, tau=1.0, a=0):
    return egh( x, H, tR, sigma, tau )

class EGHA(EGH):
    def __init__(self, **kwargs):
        func = egha if get_setting('enable_affine_tran') else egh_no_affine
        Model.__init__(self, func, **kwargs)

    def guess(self, y, x=None, negative=False, **kwargs):

        n   = np.argmax( y )
        H   = y[n]
        yf  = H * ALPHA

        iL  = bisect_right( y[0:n], yf )
        A   = x[n] - x[iL]
        iR   = bisect_right( list(reversed(y[n:])), yf )
        B   = x[len(x)-1-iR] - x[n]

        if False:
            print( 'iL=', iL, 'A=', A )
            print( 'iR=', iR, 'B=', B )
            print( 'n=', n, 'x[n]=', x[n] )

        tR  = x[n]
        sigma = np.sqrt( -B*A/(2*LN_ALPHA) )
        tau = -(B - A)/LN_ALPHA

        return np.array([H, tR, sigma, tau, 0])

    def get_name( self ):
        return _get_model_name( self.name )

    def get_param_hints( self, pkey ):
        assert pkey == 'tau'
        return EGH_TAU_LOWER_LIMIT, EGH_TAU_UPPER_LIMIT

    def x_from_height_ratio(self, ecurve, ratio, params):
        # mu, sigma, tau, a
        return x_from_height_ratio_impl(egha, ecurve, ratio, *params[1:])

    def get_params_string(self, params):
        return 'h=%g, tR=%g, sigma=%g, tau=%g, a=%g' % tuple(params)

    def guess_a_peak_with_prop(self, x, y, prop):
        """
        This function is used to guess a peak with a given area ratio,
        which has been generated (after the implementation of the same method of EGH)
        completely by Cody AI.
        """
        H, tR, sigma, tau, a = self.guess(y, x=x)
        total_area = np.sum(y)
        def objective(p):
            sigma_, tau_, a_  = p
            cy = egha(x, H, tR, sigma_, tau_, a_)
            area = np.sum(cy)
            return abs(area - total_area*prop)

        ret = minimize(objective, [sigma, tau, a], method='Nelder-Mead')
        sigma, tau, a = ret.x
        return np.array([H, tR, sigma, tau, a])
    
    def guess_binary_peaks(self, x, y, p1, p2, guess_tau=False, debug=False, plot_info=None):
        scy = np.cumsum(y)
        p1y = scy[-1]*p1
        j = bisect_right(scy, p1y)
        params_list = []
        spline = UnivariateSpline(x, y, s=0, ext=3)
        for x_, y_ in [(x[:j],y[:j]), (x[j:],y[j:])]:
            M = compute_moments(x_, y_)
            if guess_tau:
                # this guess is not so good as expected
                tR, sigma, tau = compute_egh_params([M[0], np.sqrt(M[1]), 0], M)
            else:
                # this seems better for this purpose
                tR, sigma, tau = M[0], np.sqrt(M[1]), 0
            h = spline(tR)
            params_list.append( [h, tR, sigma, tau, 0] )
        params_array = np.array(params_list)

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title('EGHA guess_binary_peaks: p1=%g, p2=%g' % (p1, p2), fontsize=16)
                if plot_info is not None:
                    ax.plot(*plot_info, color='gray', alpha=0.3)
                ax.plot(x, y)
                axt = ax.twinx()
                axt.grid(False)
                axt.plot(x, scy, ':')
                axt.plot(x[j], p1y, 'ro')           
                for params in params_array:
                    cy = egha(x, *params)
                    ax.plot(x, cy, '--')
                    ax.axvline(params[1], color='red')
                fig.tight_layout()
                plt.show()

        return params_array


"""
    Models/RateTheory/EdmParamInspect.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar


class EdmParamInspect(Dialog):
    def __init__(self, parent, dialog):
        self.parent = parent
        self.dialog = dialog
        Dialog.__init__(self, parent, "Edm Parameter Inspection", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        bframe = Tk.Frame(body_frame)
        bframe.pack(fill=Tk.X, padx=20)

        tframe = Tk.Frame(bframe)
        tframe.pack(side=Tk.LEFT)
        pframe = Tk.Frame(bframe, width=100)
        pframe.pack(side=Tk.RIGHT)

        fig, ax = plt.subplots()
        self.fig = fig
        fig.tight_layout()
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

def demo(caller):
    dialog = caller.dialog
    inspect = EdmParamInspect(dialog, dialog)
    inspect.show()



"""
    Optimizer.Devel.BoundsInspect.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import os
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

def bounds_inspect_impl(caller):
    canvas = caller.canvas
    optimizer = canvas.fullopt
    init_params = canvas.get_current_params()
    real_bounds = np.array(optimizer.get_param_bounds(init_params))
    print("real_bounds: ", real_bounds)

    with plt.Dp(button_spec=["OK", "Cancel"]):
        fig, ax = plt.subplots()
        fig.suptitle("Bounds Inspect")
        fig.tight_layout()
        plt.show()


"""
    Estimators.EghEstimator.py

    Copyright (c) 2022-2025, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.SecTheory.SecEstimator import guess_initial_secparams
from molass_legacy.Baseline.Constants import SLOPE_SCALE

if True:
    from importlib import reload
    import molass_legacy.Estimators.BaseEstimator
    reload(molass_legacy.Estimators.BaseEstimator)
from .BaseEstimator import BaseEstimator

class EghEstimator(BaseEstimator):
    def __init__(self, editor):
        BaseEstimator.__init__(self, editor)

    def estimate_egh_params(self, lrf_src=None, debug=False):
        editor = self.editor

        init_xr_params, init_xr_baseparams, init_mapping, init_uv_heights, temp_uv_baseparams = get_peak_params_advanced(editor, lrf_src=lrf_src, debug=debug)
        init_uv_baseparams = temp_uv_baseparams.copy()
        init_uv_baseparams[4:6] /=SLOPE_SCALE


        (xr_curve, D), rg_curve = editor.dsets[0:2]

        init_rgs = rg_curve.get_rgs_from_trs(init_xr_params[:,1])
        Npc, rp, tI, t0, P, m = guess_initial_secparams(init_xr_params, init_rgs)
        init_sec_params = np.array([Npc, rp, tI, t0, P, m])
    
        x = xr_curve.x
        init_mappable_range = (x[0], x[-1])

        if debug:
            self.logger.info("init_xr_params=%s", str(init_xr_params))
            self.logger.info("init_xr_baseparams=%s", str(init_xr_baseparams))
            self.logger.info("init_rgs=%s", str(init_rgs))
            self.logger.info("init_mapping=%s", str(init_mapping))
            self.logger.info("init_uv_heights=%s", str(init_uv_heights))
            self.logger.info("init_mappable_range=%s", str(init_mappable_range))
            self.logger.info("init_sec_params=%s", str(init_sec_params))

        return init_xr_params, init_xr_baseparams, init_rgs, init_mapping, init_uv_heights, init_uv_baseparams, init_mappable_range, init_sec_params

    def estimate_params(self, lrf_src=None, debug=False):
        init_xr_params, init_xr_baseparams, init_rgs, init_mapping, init_uv_heights, init_uv_baseparams, init_mappable_range, seccol_params = self.estimate_egh_params(lrf_src=lrf_src, debug=debug)
        init_params = np.concatenate([init_xr_params.flatten(), init_xr_baseparams, init_rgs, init_mapping, init_uv_heights, init_uv_baseparams, init_mappable_range, seccol_params])

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            with plt.Dp():
                fig, axes = plt.subplots(ncols=2, figsize=(12,5))
                axis_info = (fig, (*axes, None, None)) 
                self.editor.fullopt.objective_func(init_params, plot=True, axis_info=axis_info)
                fig.tight_layout()
                plt.show()

        return init_params

def get_peak_params_advanced(editor, lrf_src=None, affine=False, debug=False):
    # note that this is called from the estimators

    a, b = editor.get_pre_recog_mapping_params()
    if lrf_src is None:
        uv_peaks, xr_peaks = editor.peak_params_set[0:2]
    else:
        # consider making this a method of the LrfSource class to take into account the rg_info indeces
        uv_peaks = lrf_src.uv_peaks
        xr_peaks = lrf_src.xr_peaks

    if not affine:
        non_affine_params = []
        for params in xr_peaks:
            non_affine_params.append(params[0:4])
        xr_peaks = np.asarray(non_affine_params)

    xr_base_params = editor.baseline_params[1]
    editor.logger.info("get_peak_params_advanced: xr_base_params=%s", str(xr_base_params))

    init_mapping = a, b
    uv_heights = [peak[0] for peak in uv_peaks]
    editor.logger.info("get_peak_params_advanced: uv_heights=%s", str(uv_heights))

    uv_base_params = editor.get_uv_base_params(debug=debug)

    return np.array(xr_peaks), np.array(xr_base_params), np.array(init_mapping), np.array(uv_heights), uv_base_params



"""
    _MOLASS.AppSingleInstance.py

    Copyright (c) 2019-2024, SAXS Team, KEK-PF
"""
import sys
from molass_legacy.KekLib.SingleInstance import SingleInstance
try:
    import molass_legacy.KekLib.CustomMessageBox as MessageBox
except:
    import molass_legacy.KekLib.OurMessageBox as MessageBox

def single_instance_check(lock_path):
    si = None
    try:
        si = SingleInstance(lock_path=lock_path, raise_=False)
        if si.is_running is None:
            print('unexpected state for', lock_path)
        elif si.is_running:
            from molass_legacy.SerialAnalyzer.SerialTestUtils import get_tk_root
            root = get_tk_root()
            ret = MessageBox.askyesno("Already Running Notification",
                '_MOLASS seems to be already running!\n'
                '\n'
                'Multiple instance execution is not yet well supported.\n'
                'However, you can continue at your own risk.\n'
                '\n'
                'Would you really like to continue?'
                ,
                parent=root)
            if ret:
                pass
            else:
                sys.exit("User canceled a multiple instance execution!")
    finally:
        pass

    return si



"""
    update-custom-codes.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import sys
import os
from shutil import copy
import diff_match_patch as dmp_module

try:
    GIT_REPOSITORY = r"D:\Github\denss"
    assert os.path.exists(GIT_REPOSITORY)
except:
    GIT_REPOSITORY = r"E:\Github\denss"
    assert os.path.exists(GIT_REPOSITORY)

git_bin = os.path.join(GIT_REPOSITORY, "bin")
git_lib = os.path.join(GIT_REPOSITORY, "saxstats")

this_dir = os.path.dirname(os.path.abspath( __file__ ))
sys.path.append( this_dir + '/../../lib' )

import molass_legacy.KekLib
from DiffUtils import file2string, string2file

dmp = dmp_module.diff_match_patch()
dmp.Patch_DeleteThreshold = 0.7
dmp.Match_Distance = 5000

# saxstats
old_bin = "bin"
old_src_path = os.path.join(old_bin, "denss.fit_data.py")
old_fit_data = file2string(old_src_path)
tmp_bin = "bin-temp"
tmp_src_path = os.path.join(tmp_bin, "denss.fit_data.py")
tmp_fit_data = file2string(tmp_src_path)

def extract_func(path):
    extracted = []
    in_the_func = False
    with open(path) as fh:
        for line in fh:
            if line.find("def estimate_dmax(") >= 0:
                in_the_func = True

            if in_the_func:
                extracted.append(line)
                if line.find("return") >= 0:
                    break
    return "".join(extracted)

old_lib = "saxstats"
old_src_path = os.path.join(old_lib, "saxstats.py")
old_saxstats = extract_func(old_src_path)

tmp_lib = "saxstats-temp"
tmp_src_path = os.path.join(tmp_lib, "saxstats.py")
tmp_saxstats = extract_func(tmp_src_path)

k = 0
for old_src, file, tmp_src in [
                        (old_fit_data, "DenssUtils.py", tmp_fit_data), 
                        (old_saxstats, "DmaxEstimation.py", tmp_saxstats)]:

    mod_src = file2string(file)
    patches = dmp.patch_make(old_src, mod_src)
    print(dmp.patch_toText(patches))

    results = dmp.patch_apply(patches, tmp_src)

    new_src = results[0]
    print("=====================", results[1])
    assert results[1] == [True] * len(results[1])

    temp_file = file.replace(".py", "-temp.py")
    # string2file(new_src, temp_file)
    with open(temp_file, "w", newline="\n") as fh:
        fh.write(new_src)



# coding: utf-8
"""

    LinearModel.py

"""
import numpy            as np
import scipy.special    as sp

def linear_model(B, x):
    '''Linear function y = m*x + b'''
    # B is a vector of the parameters.
    # x is an array of the current x values.
    # x is in the same format as the x passed to Data or RealData.
    #
    # Return an array in the same format as y passed to Data or RealData.
    return B[0]*x + B[1]



"""
    Optimizer.OptConstants.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
MIN_NUM_PEAKS = 1
MAX_NUM_PEAKS = 7
MIN_NUM_COMPONENTS = MIN_NUM_PEAKS      # include cases where wihout baseline component
MAX_NUM_COMPONENTS = MAX_NUM_PEAKS + 1



# coding: utf-8
"""

    ScatteringBasecurve.py

    scattering baseline solver

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF

"""
import numpy            as np
from lmfit              import minimize, Parameters
from scipy              import stats
import OurStatsModels   as sm

DEBUG_PLOT              = False
PERCENTILE_FIRST        = 30
PERCENTILE_SECOND       = 25
PERCENTILE_FINAL        = 20
NEGATIVE_VALUE_PENALTY  = 10

class ScatteringBasecurve:
    def __init__( self, y ):
        self.y  = y
        self.x = np.arange( len(self.y) )

    def solve( self, p_final=PERCENTILE_FINAL ):
        xpp, ypp, A, B, C, ppp = self.get_low_percentile_params( self.x, self.y, p_final )

        b_  = A*self.x**2 + B*self.x + C
        y_  = self.y - b_
        yf  = np.percentile( y_, p_final )
        p_  = np.where( y_ <= yf )[0]
        n   = np.argmax( y_[p_] )
        xf  = p_[n]
        C_  = self.y[xf] - ( A*xf**2 + B*xf )

        self.xpp    = xpp
        self.ypp    = ypp
        self.npp    = n
        self.params = ( A, B, C_ )
        return A, B, C_

    def solve_quadratic( self, xpp, ypp ):
        """
            y = A * X**2 + B * x + C

            X = [   [ 0**2, 0, 1 ],
                    [ 1**2, 1, 1 ],
                    [ 2**2, 2, 1 ],
                    ...
                    [ (n-1)**2, (n-1), 1 ],
                ]
        """
        Xpp = np.array( [ xpp**2, xpp, np.ones( len(xpp) ) ] ).T
        model   = sm.OLS( ypp, Xpp )
        result  = model.fit()
        A, B, C = result.params
        return A, B, C

    def get_low_percentile_params( self, x, y, p_final ):
        ppp = np.percentile( y, [ PERCENTILE_FIRST ] )
        xpp = np.where( y <= ppp[0] )[0]
        ypp = y[xpp]
        n   = 2

        for i in range(n):
            A, B, C = self.solve_quadratic( xpp, ypp )
            y_ = y - ( A*x**2 + B*x + C )
            ppp = np.percentile( y_, [ PERCENTILE_SECOND, p_final ] )
            if i == n - 1:
                break

            xpp = np.where( y_ <= ppp[0] )[0]
            ypp = y[xpp]

        return xpp, ypp, A, B, C, ppp

    def debug_plot( self, title="Debug", parent=None ):
        from DebugCanvas    import DebugCanvas
        A, B, C = self.params

        def debug_plot( fig ):
            ax = fig.add_subplot( 111 )

            ax.set_title( title )
            ax.plot( self.y )
            # ax.plot( self.y, ':', color='green' )
            ax.plot( self.xpp, self.ypp, 'o', color='yellow', label='under 25% points' )


            baseline_ = A*self.x**2 + B*self.x + C
            ax.plot( baseline_, color='red', alpha=0.5, label='low percentile linear' )
            ax.plot( self.npp, self.y[self.npp], 'o', color='red', label='10% point' )

            ax.legend()
            fig.tight_layout()

        dc = DebugCanvas( "Debug", debug_plot, parent=parent, toolbar=True )
        dc.show( cancelable=True )
        return dc.continue_



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import mpl_toolkits.mplot3d.art3d as art3d
import matplotlib.animation as animation

x = np.linspace(0, 10*np.pi, 100)
y = np.sin(x)
fig, (ax1, ax2) = plt.subplots(nrows=2)
line, = ax1.plot(x, y)

ax2.set_ylim(0,1)
z = np.ones(len(x))*0.5
artists = ax2.scatter(x, z, vmin=-1, vmax=1, c=y, cmap=cm.plasma)

def init():
  line.set_data([],[])
  artists.set_array(y)
  return [line, artists]

num_frames = 40

def animate(i):
  y = np.sin(x-10*np.pi*i/num_frames)
  line.set_data(x,y)
  artists.set_array(y)
  return [line, artists]

anim = animation.FuncAnimation(fig, animate, init_func=init,
                                frames=num_frames, interval=100, blit=True)
# anim.save('beam_wave.mp4', writer="ffmpeg")
plt.show()



"""
    UV.UvBaseSpline.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from scipy.interpolate import UnivariateSpline
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Trimming.Sigmoid import guess_bent_sigmoid, ex_sigmoid, adjust_ex_sigmoid
from molass_legacy.Peaks.EghSupples import egh, d_egh
from molass_legacy.KekLib.SciPyCookbook import smooth
from molass_legacy._MOLASS.SerialSettings import get_setting
from .Constants import SLOPE_SCALE

SMOOTHING = True
PENALTY_SCALE = 1e8
CERTAINTY_LIMIT = 0.5
BETTER_LOWER_NEEDED_RATIO = 0.5
TRY_ADJUST_EX_SIGMOID = True

class UvBaseSpline:
    def __init__(self, sd, curve1=None, spline=None):
        self.sd = sd
        self.curve1 = curve1
        if spline is None:
            spline = curve1.d1
        self.diff_spline = spline
        self.baseline_type = get_setting("unified_baseline_type")
        self.lower_xy = None
        self.base_xy = None

    def guess_params(self, curve2, pp, slice_, fc_point=None, debug=False):
        x = curve2.x
        y = curve2.y

        x_ = x[slice_]
        if SMOOTHING:
            y_ = smooth(y[slice_])
        else:
            y_ = y[slice_]

        certainty = curve2.sigmoid_certainty
        if certainty < CERTAINTY_LIMIT:
            sigm_params = None
            if debug:
                print("(1) certainty=", certainty)
            if certainty == 0:
                if debug:
                    from importlib import reload
                    import Trimming.SigmoidApplicability
                    reload(Trimming.SigmoidApplicability)
                from molass_legacy.Trimming.SigmoidApplicability import check_applicability
                ok, info = check_applicability(x_, y_, debug=debug)
                if not ok:
                    sigm_params = list(info.safe_params)
            if sigm_params is None:
                sigm_params = list(guess_bent_sigmoid(x_, y_, debug=debug))
        else:
            sigm_params = list(curve2.sigmoid_params)

        if debug:
            print("(2) certainty=", certainty, CERTAINTY_LIMIT)
            print("fc_point=", fc_point)
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("guess_params (1)")
                ax.plot(x, y, label="data")
                ax.plot(x_, y_, label="smoothed")
                ax.plot(x, ex_sigmoid(x, *sigm_params), label="ex_sigmoid")
                ax.legend()
                fig.tight_layout()
                plt.show()

        if fc_point is None:
            force_L0_x0 = True
            sigm_params[0:2] = (0,0)    # this modification should be considerd earlier
        else:
            force_L0_x0 = False         

        ty = None       # dummy arg which will not be used here

        def obj_func(p, debug=False):
            fy = self.__call__(x_, p, ty)
            fv = np.sum((fy - y_)**2)
            if debug:
                print("fv=", fv)
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.plot(x_, y_)
                    ax.plot(x_, fy)
                    fig.tight_layout()
                    plt.show()
            if force_L0_x0:
                # or consider using bounds
                fv += PENALTY_SCALE*np.sum(p[0:2]**2)   # L, x0 penalty
            return fv

        opt_ret = None
        for scale in [-1, 1]:
            init_params = (*sigm_params, scale)
            ret = minimize(obj_func, init_params, method='Nelder-Mead')
            # ret = basinhopping(obj_func, init_params)
            print("ret.x=", ret.x)
            if opt_ret is None or ret.fun < opt_ret.fun:
                opt_ret = ret

        print("opt_ret.fun=", opt_ret.fun)

        x1 = self.curve1.x
        y1 = self.curve1.y
        # use lower part only
        lower = y1 < 0.1* self.curve1.max_y
        x1_ = x1[lower]
        y1_ = y1[lower]
        lower_ratio = len(x1_)/len(x1)
        if debug:
            print("lower_ratio=", lower_ratio)
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("lower part check")
                ax.plot(x1, y1, label="data")
                ax.plot(x1_, y1_, ":", label="lower part")
                ax.legend()
                fig.tight_layout()
                plt.show()

        if TRY_ADJUST_EX_SIGMOID:
            if lower_ratio < BETTER_LOWER_NEEDED_RATIO:
                # as in 20191118_4
                x1_, y1_ = self.make_better_lower_part(self.curve1)
            adjusted_params = adjust_ex_sigmoid(x1_, y1_, opt_ret.x)
        else:
            adjusted_params = opt_ret.x

        if debug:
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                ax1.set_title("sliced data")
                ax1.plot(x, y, label="data")
                ax1.plot(x_, y_, label="sliced smooth data")
                ax1.legend()
                ax2.set_title("after adjust_ex_sigmoid")
                ax2.plot(x1, y1, label="data")
                ax2.plot(x1_, y1_, ":", label="sliced data")
                ax2.plot(x1, ex_sigmoid(x1, *opt_ret.x[:-1]), label="optimized params")
                ax2.plot(x1, ex_sigmoid(x1, *adjusted_params[:-1]), label="adjusted params")
                ax2.legend()
                fig.tight_layout()
                plt.show()

        self.lower_xy = x1_, y1_
        if fc_point is None:
            self.base_xy = self.lower_xy
        else:
            # as in 20171203
            extended_base = np.logical_or(lower, x <= fc_point)
            self.base_xy = x[extended_base], y[extended_base]

        if debug:
            from molass_legacy.UV.PlainCurveUtils import get_both_wavelengths

            sd = self.sd
            wavelengths = get_both_wavelengths(sd.lvector)
            with plt.Dp():
                from bisect import bisect_right
                from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
                from molass_legacy.Elution.CurveUtils import simple_plot
                from molass_legacy.DataStructure.MatrixData import simple_plot_3d
                fig = plt.figure(figsize=(18,10))
                fig.suptitle("guess_params (2) for %s" % get_in_folder(), fontsize=20)
                ax1 = fig.add_subplot(221)
                ax2 = fig.add_subplot(222)
                ax3 = fig.add_subplot(223, projection="3d")
                ax4 = fig.add_subplot(224)

                ax1.set_title(r"Elution at $\lambda=%.3g$" % wavelengths[0], fontsize=16)
                simple_plot(ax1, self.curve1, legend=False)
                by280 = self.__call__(x, adjusted_params, ty)
                by400 = self.__call__(x, opt_ret.x, ty)
                ax1.plot(x, by280, color="red", label="baseline at 280")
                ax1.plot(x, by400, ":", color="red", label="baseline at 400")
                ax1.legend(fontsize=16)
                ax1.set_xlabel("Eno")

                ax2.set_title(r"Elution at $\lambda=%.3g$" % wavelengths[1], fontsize=16)
                ax2.plot(x, y)
                if SMOOTHING:
                    ax2.plot(x_, y_)
                fy = self.__call__(x, opt_ret.x, ty)
                ax2.plot(x, fy)
                ax2.set_xlabel("Eno")

                ax3.set_title("3D View", fontsize=16)
                ax4.set_title("Complementary View", fontsize=16)
                ax4.set_xlabel(r"$\lambda$")
                wv = sd.lvector
                U = sd.conc_array
                simple_plot_3d(ax3, U, x=wv)
                ones = np.ones(len(wv))
                for j, color in (len(x)-1, "cyan"), (0, "pink"):
                    z = U[:,j]
                    ax3.plot(wv, ones*j, z, color=color)
                    ax4.plot(wv, z, color=color, label="j = %d" % j)

                ymin, ymax = ax4.get_ylim()
                ax4.set_ylim(ymin, ymax)
                ones = np.ones(len(x))
                jj = np.arange(len(x))

                for w in wavelengths:
                    i = bisect_right(wv, w)
                    ax3.plot(ones*w, jj, U[i,:], color="yellow")
                    ax4.plot([w, w], [ymin, ymax], label=r"$\lambda=%.3g$" % w, color="yellow")

                ax4.legend(fontsize=16)
                fig.tight_layout()
                plt.show()

        if self.baseline_type == 1:
            ret_params = adjusted_params
        elif self.baseline_type == 2:
            intreg_scale = 0
            ret_params = np.concatenate([adjusted_params, [intreg_scale]])
        elif self.baseline_type == 3:
            intreg_scale = 0
            ret_params = np.concatenate([adjusted_params, [intreg_scale]])
        else:
            assert False

        ret_params[4:6] *= SLOPE_SCALE
        return ret_params

    def guess_integral_scale(self, x, y, params, ty=None, debug=False):
        assert len(params) == 7
        if ty is None:
            by = self.__call__(x, params, y)    # y won't be used
            ty = y - by

        opt_params = np.concatenate([params, [0]])

        def objective(p):
            opt_params[-1] = p[0]
            integ_base = self.__call__(x, opt_params, ty)   # y will be used
            return (integ_base[-1] - y[-1])**2 + PENALTY_SCALE*min(0, opt_params[-1])**2

        ret = minimize(objective, (0,))
        ret.x[0] = max(0, ret.x[0])     # force it to be non-negative

        if debug:
            opt_params[-1] = ret.x[0]
            integ_base = self.__call__(x, opt_params, ty) 
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("guess_integral_scale")
                ax.plot(x, y)
                ax.plot(x, integ_base, color="red")
                fig.tight_layout()
                plt.show()

        return ret.x[0], ty

    def make_better_lower_part(self, curve, debug=False):
        from molass_legacy.Baseline.ScatteringBaseline import ScatteringBaseline
        x = curve.x
        y = curve.y
        sbl = ScatteringBaseline(y, x=x)
        A, B = sbl.solve()
        y_ = A*x+B
        if curve.is_roughly_centered() and abs(y_[-1] - y[-1])/curve.max_y > 0.05:
            # as in 20160628
            # this works, but more general solution is desirable
            # print("retrying make_better_lower_part")
            A, B = sbl.solve(p_final=50)
            y_ = A*x+B        
        if debug:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("make_better_lower_part")
                ax.plot(x, y)
                ax.plot(x, y_, color="red")
                fig.tight_layout()
                plt.show()
        return x, y_

    def __call__(self, x, arg_params, ty, cy_list=None):
        params = arg_params.copy()
        params[4:6] /= SLOPE_SCALE
        k = 7
        y = compute_baseline_impl(x, params[0:k], self.diff_spline)
        if len(params) == 8:
            if cy_list is None:
                y += params[k]*np.cumsum(ty)
            else:
                for cy in cy_list:
                    y += params[k]*np.cumsum(cy)
        return y
           
    def get_shifted(self, n):
        x = self.curve1.x
        dy = self.diff_spline(x)
        shifted_spline = UnivariateSpline(x-n, dy, s=0, ext=3)
        return UvBaseSpline(self.sd, spline=shifted_spline)

def compute_baseline_impl(x, params, diff_spline):
    """
        requires 6 + 1 params
    """
    y = np.zeros(len(x))
    y += ex_sigmoid(x, *params[0:-1])
    y += params[-1]*diff_spline(x)
    return y


# coding: utf-8
"""
    DecompProcessUtils.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import os
import numpy as np

MAXNUM_STEPS = 20000

def run_decomp_impl(*args, **kwargs):
    pass

def get_decompfolder():
    from molass_legacy._MOLASS.SerialSettings import get_setting
    from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry

    decomp_folder = (get_setting('analysis_folder') + '/Decomp').replace('\\', '/')
    if not os.path.exists(decomp_folder):
        mkdirs_with_retry(decomp_folder)

    return decomp_folder

def get_outfolder(job_id=None):
    decomp_folder = get_decompfolder()

    import re
    if job_id is None:
        out_folder = decomp_folder + '/000'
        while True:
            if os.path.exists(out_folder):
                out_folder = re.sub(r'/(\d+)$', lambda m: '/%03d' % (int(m.group(1)) + 1), out_folder)
            else:
                break
    else:
        out_folder_init = decomp_folder + '/%03d' % job_id
        out_folder = out_folder_init
        i = 0
        while True:
            if os.path.exists(out_folder):
                i += 1
                out_folder =  out_folder_init + '_%d' % i
            else:
                break

    return out_folder



# coding: utf-8
"""
    ResultPrinter.py

    Copyright (c) 2016-2018, Masatsuyo Takahashi, KEK-PF
"""
import numpy                as np

class ResultPrinter:
    def __init__( self, fh, csv_output, ATSAS=False ):
        self.fh     = fh
        self.format = None
        self.csv_output = csv_output
        self.separator = ',' if self.csv_output else '\t'
        self.ATSAS  = ATSAS

    def print_result( self, i, file, result, multiple=False ):
        file_ = file.replace( '/', '\\' )

        if multiple:
            self.print_result_multiple( i, file_, result )
        else:
            self.print_result_single( i, file_, result )

    def print_result_single( self, i, file, result ):
        if self.format is None and not self.csv_output:
            stdev_p = round( result.Rg_stdev / result.Rg * 100 )
            print( "Rg   = %.1f +/- %.2g (%d%%)" % ( result.Rg, result.Rg_stdev, stdev_p ) )
            print( "I(0) = %.3g +/- %.2g" % ( result.I0, result.I0_stdev ) )
            print( "Points %d to %d (%d total)" % ( result.From, result.To, result.To - result.From + 1 ) )
            print( "Quality %.2f" % result.Quality )
        else:
            self.print_result_multiple( i, file, result )

    def print_result_multiple( self, i, file, result ):
        if i == 0:
            self.print_header_line()
        self.print_detail_line( file, result )

    def print_header_line( self ):
        columns =  [ 'File',
                    'Rg', 'Rg StDev',
                    'I(0)', 'I(0) StDev',
                    'First point', 'Last point',
                    'Quality', 'Aggregated' ]

        if not self.ATSAS:
            columns += [ 'Min Q', 'Max Q', 'Remarks' ]
            # columns += [ 'Mono G', 'Mono Rg', 'Mono d' ]
            # columns += [ 'QF basic quality', 'QF positive score', 'QF fit consistency', 'QF stdev score', 'QF qRg score' ]
            # columns += [ 'IpI flag', 'BiCo flag', 'b/m ratio' ]
            # columns += [ 'BiCo G1', 'BiCo G2', 'BiCo Rg1', 'BiCo Rg2', 'BiCo d1', 'BiCo d2' ]

        self.fh.write( self.separator.join( columns ) + '\n' )

    def print_detail_line( self, file, result ):

        aggregated_ = '0'

        str_vars = []
        for v in [result.Rg, result.Rg_stdev, result.I0, result.I0_stdev,]:
            if v is None or np.isnan( v ):
                s_ = 'NA'
            else:
                s_ = '%g' % v
            str_vars.append( s_ )
        for v in [ result.From, result.To, ]:
            str_vars.append( 'NA' if v is None else '%d' % v )

        quality_ = 'NA' if result.Quality is None else '%.2g' % result.Quality 

        self.fh.write( ','.join( [  file, *str_vars, quality_, aggregated_ ] ) )

        if not self.ATSAS:
            # 
            str_vars = map( lambda x: 'NA' if x  is None else '%g' % x, [ result.min_q, result.max_q ] )
            self.fh.write( ',%s,%s,"%s"' % ( *str_vars, 'Unused' ) )

        if False:
            # Mono
            fit = result.fit
            str_vars = map( lambda x: 'NA' if x  is None else '%g' % x, [ fit.I0, fit.Rg, fit.degree ] )
            self.fh.write( ',' + ','.join( str_vars ) )

            # QF
            qobj = result.quality_object
            if qobj is None:
                str_vars = [ 'NA' ] * 5
            else:
                str_vars = map( lambda x: 'NA' if x  is None else '%g' % x, [ qobj.basic_quality, qobj.positive_score, qobj.fit_consistency_pure, qobj.stdev_score, qobj.q_rg_score ] )
            self.fh.write( ',' + ','.join( str_vars ) )

            # flags
            str_vars = map( lambda x: 'NA' if x  is None else '%g' % x, [ result.IpI, result.bicomponent, result.bico_mono_ratio ] )
            self.fh.write( ',' + ','.join( str_vars ) )

            # BiCo
            bico_params = [ result.bico_G1, result.bico_G2, result.bico_Rg1, result.bico_Rg2, result.bico_d1, result.bico_d2 ]
            str_vars = map( lambda x: 'NA' if x  is None else '%g' % x, bico_params )
            self.fh.write( ',' + ','.join( str_vars ) )

        self.fh.write( '\n' )



"""
    Models.Stochastic.SDM.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.Models.Tentative import Model
from molass_legacy.Models.Stochastic.DispersivePdf import dispersive_monopore_pdf
from molass_legacy.Models.ElutionModelUtils import x_from_height_ratio_impl

class SDM(Model):
    def __init__(self, **kwargs):
        super(SDM, self).__init__(dispersive_monopore_pdf, **kwargs)
        self.fx = None

    def get_name(self):
        return "SDM"

    def is_traditional(self):
        return False

    def eval(self, params=None, x=None):
        return self.func(x, *params)
    
    def x_from_height_ratio(self, ecurve, ratio, params):
        assert self.fx is not None
        return x_from_height_ratio_impl(dispersive_monopore_pdf, ecurve, ratio, *params, needs_ymax=True, full_params=True, fx=self.fx)
    
    def set_fx_for_height_ratio(self, fx):
        self.fx = fx

    def get_params_string(self, params):
        return 'scale=%g, N=%g, T=%g, t0=%g, me=%s, mp=%g, rp=%g, rg=%g' % tuple(params)


# coding: utf-8
"""
    WeightedLRF.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import copy
import numpy as np

SMALL_POSITIVE_VALUE = 1e-10

def compute_reciprocally_weighted_matrices(conc_dependence, C, D):
    assert C.shape[0] <= 2 and C.shape[0] == conc_dependence
    assert C.shape[1] == D.shape[1]

    row_ones = np.ones(C.shape[1])
    n = C.shape[0]//conc_dependence
    c = copy.deepcopy(C[0,:])
    c[c < SMALL_POSITIVE_VALUE] = SMALL_POSITIVE_VALUE

    c_rows = [row_ones]
    if conc_dependence > 1:
        c_rows.append(c)
    C_ = np.array(c_rows)

    cinv = 1/c
    d_rows = []
    for i in range(D.shape[0]):
        d_rows.append(D[i,:]*cinv)

    D_ = np.array(d_rows)
    return C_, D_



"""
    Basseline.Inspect.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Batch.LiteBatch import LiteBatch

def baseline_inspect_impl(caller):
    print('baseline_inspect_impl')
    
    lb = LiteBatch()
    lb.prepare(caller.serial_data, debug=True)
    uv_x, uv_y, xr_x, xr_y, baselines = lb.get_curve_xy(return_baselines=True, debug=False)
    uv_y_ = uv_y - baselines[0]
    xr_y_ = xr_y - baselines[1]
    # uv_peaks, xr_peaks = lb.get_modeled_peaks(uv_x, uv_y_, xr_x, xr_y_)

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 5))
        fig.suptitle("Baseline inspection Result")
        ax1.set_title("UV")
        ax1.plot(uv_x, uv_y)
        ax1.plot(uv_x, baselines[0], color="red")

        ax2.set_title("XR")
        ax2.plot(xr_x, xr_y)
        ax2.plot(xr_x, baselines[1], color="red")
        
        fig.tight_layout()
        plt.show()


# -*- coding: utf-8 -*-
"""

    ファイル名：   CreateFolderDialog.py

    処理内容：

        GUI のオプション設定情報

"""
from __future__ import division, print_function, unicode_literals

import sys
import os

from molass_legacy.KekLib.OurTkinter     import Tk, Dialog
from molass_legacy.KekLib.TkSupplements  import tk_set_icon_portable

class CreateFolderDialog( Dialog ):

    def __init__( self, parent, title, suggested_folders ):
        self.grab = 'local'     # used in grab_set
        self.suggested_folders  = suggested_folders
        self.applied    = False
        self.created_folder = None

        Dialog.__init__(self, parent, title) # this calls body

    def body( self, body_frame ):   # overrides parent class method

        tk_set_icon_portable( self, 'synthesizer' )

        iframe = Tk.Frame( body_frame );
        iframe.pack( expand=1, fill=Tk.BOTH, padx=10, pady=10 )

        question_label = Tk.Label( iframe, text= 'Select (and change if necessary) one of the new folders suggested below to create.' )
        question_label.pack( anchor=Tk.W )

        entry_frame = Tk.Frame( iframe );
        # entry_frame.pack( expand=1, fill=Tk.BOTH )
        entry_frame.pack( anchor=Tk.W )

        text_entry_width = 80
        self.folders = []

        self.selection = Tk.IntVar()
        self.selection.set( 0 )
        i = 0
        for f in self.suggested_folders:
            b = Tk.Radiobutton( entry_frame, text='', variable=self.selection, value=i )
            b.grid( row=i, column=0, sticky=Tk.W )
            var = Tk.StringVar()
            var.set( f )
            self.folders.append( var )
            entry_ = Tk.Entry( entry_frame, textvariable=var, width=text_entry_width )
            entry_.grid( row=i, column=1, sticky=Tk.W )
            i += 1

        guide_label = Tk.Label( iframe, text= 'If you cancel, the file selection dialog will follow.' )
        guide_label.pack( anchor=Tk.W )

        # global grab cannot be set befor windows is 'viewable'
        # and this happen in mainloop after this function returns
        # Thus, it is needed to delay grab setting of an interval
        # long enough to make sure that the window has been made
        # 'viewable'
        if self.grab == 'global':
            self.after(100, self.grab_set_global )
        else:
            pass # local grab is set by parent class constructor

    def apply( self ):  # overrides parent class method
        print( "CreateFolderDialog: ok. apply" )

        i = self.selection.get()
        selected_folder =  self.folders[i].get()
        if not os.path.exists( selected_folder ):
            os.makedirs( selected_folder )
        self.created_folder = selected_folder
        self.applied    = True



# coding: utf-8
"""

    ファイル名：   PilatusUtils.py

    処理内容：

       Pilatus 画像データ関連の共通処理

    Copyright (c) 2020, 2025, SAXS Team, KEK-PF
"""
import os
import re
import glob
import PilatusCounter
from PilatusUtilsOldStyle import get_data_info as get_data_info_old_style
from PilatusUtilsNewStyle import get_data_info as get_data_info_new_style, file_name_re

DEBUG = False
NEW_STYLE_ONLY = True

def get_in_folder_info( in_folder ):
    orig_folder = os.getcwd()
    os.chdir( in_folder )
    log_files = glob.glob( 'measure*.log' )
    if len( log_files ) == 0:
        log_files = glob.glob( '*.log' )
    mask_files = glob.glob( '*.mask' )
    os.chdir( orig_folder )

    log_file, mask_file = (None, None )
    if len( log_files ) >= 1:
        log_file = log_files[0]

    mask_file = None
    if len( mask_files ) >= 1:
        mask_file = mask_files[0]

    return log_file, mask_file

old_file_name_re = re.compile(r'(\w+)_(\d)_(\d{5})\.(\w{3})')

class FolderInfo:
    def __init__(self, path):
        self.path = path
        self.single_image = True
        self.new_style = None
        self.get_image_file_names(path)

    def get_image_file_names(self, folder):
        for ext in ['tif', 'cbf']:
            self.image_files = []
            old_type_files = []
            for k, path in enumerate(glob.glob(folder + r'\*.' + ext)):
                # print([k], path)
                folder, file= os.path.split(path)
                m = file_name_re.match(file)
                if m:
                    # print('match', [k], path)
                    self.image_files.append(path)
                    if int(m.group(3)) > 0:
                        self.single_image = False
                else:
                    if old_file_name_re.match(file):
                        old_type_files.append(path)

            if len(self.image_files) > 0 and len(old_type_files) == 0:
                if self.new_style is None:
                    self.new_style = True
                break
            else:
                if len(old_type_files) > 0:
                    self.new_style = False
                    break

        if DEBUG:
            print('-------- old_type_files=', old_type_files)
            print('-------- new_style=', self.new_style)

    def is_new_style(self):
        return self.new_style

def get_data_info( in_folder,
                    adj_folder, syn_folder, pilatus_counter, counter_id,
                    log_file_path=None,
                    sample_complete=False, for_test_data=False,
                    logger=None ):

    log_file        = None
    mask_file       = None

    if not in_folder or not os.path.exists( in_folder ):
        return log_file, mask_file, [], pilatus_counter

    log_file, mask_file = get_in_folder_info( in_folder )

    # TODO: refactor to move PilatusCounter out.
    if pilatus_counter is None:
        pilatus_counter = PilatusCounter.Counter( in_folder )

    if log_file_path is None:
        log_file_path = in_folder + '/' + log_file 

    folder_info = FolderInfo(in_folder)

    if NEW_STYLE_ONLY or folder_info.is_new_style():
        get_data_info_impl = get_data_info_new_style
    else:
        get_data_info_impl = get_data_info_old_style

    return get_data_info_impl(folder_info, log_file, mask_file,
                    adj_folder, syn_folder, pilatus_counter, counter_id,
                    log_file_path, sample_complete, for_test_data,
                    logger=logger)



"""
    SecTheory.PlateTheory.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
import io
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Rectangle
import matplotlib.image as mpimg
import matplotlib.animation as animation
import math

NUM_PLATES = 64
NUM_PARTICLES = 2000
NUM_CYCLES = 200

def update_states(plate_states, p):
    q = 1 - p
    previos_states = plate_states.copy()
    for k in range(NUM_PLATES):
        increase = 0 if k == 0 else p*previos_states[k-1]
        decrease = q*previos_states[k]
        plate_states[k] = increase + decrease

def draw_column(ax, w, h, states_list):
    dh = h/NUM_PLATES

    for k in range(NUM_PLATES):
        y = dh*(NUM_PLATES - 1 - k)
        ax.add_patch(Rectangle(xy=(0, y), width=w,  height=dh, alpha=0.5))

        for plate_states, color, msize in states_list:
            n = int(plate_states[k])
            pos = np.random.uniform(0, 1, (n, 2))
            ax.plot(w*pos[:,0], y + dh*pos[:,1], "o", color=color, markersize=msize)

def demo(save=False):

    plate_states = np.zeros(NUM_PLATES)
    plate_states[0] = NUM_PARTICLES
    p = 0.8*NUM_PLATES/100

    fig = plt.figure(figsize=(9,5))
    fig.suptitle("Plate Theory Simulation N=%d" % NUM_PLATES, fontsize=20)
    fig.tight_layout()
    fig.subplots_adjust(top=0.85, wspace=0.5)

    gs = GridSpec(1,3)

    ax1 = fig.add_subplot(gs[0,0])
    ax2 = fig.add_subplot(gs[0,1:3])

    ax2.set_title("Concentration at the End of Column")

    images = []

    el_x = []
    el_y = []
    line, = ax2.plot(el_x, el_y, lw=5)

    ax2.set_xlim(-10, NUM_CYCLES+10)
    ax2.set_ylim(-0.01, 0.11)

    for k in range(NUM_CYCLES):
        update_states(plate_states, p)

        ax1.cla()
        ax1.set_axis_off()
        ax1.set_title("Column State")   # cleared by ax.cla()

        draw_column(ax1, 1, 1, [(plate_states, "red", 1)])

        el_x.append(k)
        el_y.append(plate_states[-1]/NUM_PARTICLES)
        line.set_xdata(el_x)
        line.set_ydata(el_y)

        plt.pause(0.1)

        buffer = io.BytesIO()
        fig.savefig(buffer)
        buffer.seek(0)
        img = mpimg.imread(buffer)
        images.append(img)

    plt.show()
    plt.close()

    if not save:
        return

    fig, ax = plt.subplots(figsize=(9,5))
    ax.set_axis_off()
    fig.tight_layout()

    ims = []
    for img in images:
        im = ax.imshow(img, animated=True)
        ims.append([im])

    ani = animation.ArtistAnimation(fig, ims, interval=100, blit=True, repeat_delay=1000)
    ani.save("plate_theory.gif")

    plt.show()

def demo2(save=False):

    plate_states1 = np.zeros(NUM_PLATES)
    plate_states2 = np.zeros(NUM_PLATES)
    num_particles1 = NUM_PARTICLES//2
    num_particles2 = NUM_PARTICLES//2
    plate_states1[0] = num_particles1
    plate_states2[0] = num_particles2

    p1 = 0.9*NUM_PLATES/100
    p2 = 0.7*NUM_PLATES/100

    fig = plt.figure(figsize=(9,5))
    fig.suptitle("Plate Theory Simulation N=%d with Two Species" % NUM_PLATES, fontsize=20)
    fig.tight_layout()
    fig.subplots_adjust(top=0.85, wspace=0.5)

    gs = GridSpec(1,3)

    ax1 = fig.add_subplot(gs[0,0])
    ax2 = fig.add_subplot(gs[0,1:3])

    ax2.set_title("Concentration at the End of Column")

    images = []

    el_x = []
    el_y = []
    line, = ax2.plot(el_x, el_y, lw=5)

    ax2.set_xlim(-10, NUM_CYCLES+10)
    ax2.set_ylim(-0.01, 0.11)

    for k in range(NUM_CYCLES):
        update_states(plate_states1, p1)
        update_states(plate_states2, p2)

        ax1.cla()
        ax1.set_axis_off()
        ax1.set_title("Column State")   # cleared by ax.cla()

        draw_column(ax1, 1, 1, [(plate_states1, "red", 2), (plate_states2, "green", 1)])

        el_x.append(k)

        y = plate_states1[-1]/num_particles1 + plate_states2[-1]/num_particles2
        el_y.append(y)
        line.set_xdata(el_x)
        line.set_ydata(el_y)

        plt.pause(0.1)

        buffer = io.BytesIO()
        fig.savefig(buffer)
        buffer.seek(0)
        img = mpimg.imread(buffer)
        images.append(img)

    plt.show()
    plt.close()

    if not save:
        return

    print("saving the animation ...")

    fig, ax = plt.subplots(figsize=(9,5))
    ax.set_axis_off()
    fig.tight_layout()

    ims = []
    for img in images:
        print(".")
        im = ax.imshow(img, animated=True)
        ims.append([im])

    ani = animation.ArtistAnimation(fig, ims, interval=100, blit=True, repeat_delay=1000)
    ani.save("plate_theory.gif")

    plt.show()

def demo3():
    fig, ax = plt.subplots(figsize=(15,4))

    fig.suptitle("Pascal's Triangle in Theoretical Plates", fontsize=16)

    h = 1
    num_plates = 5
    dh = h/num_plates
    w = dh*0.7

    def term(n, k):
        assert n > 0
        m = n - k
        t = r"%d \cdot " % math.comb(n, k)
        if m > 0:
            t += "q"
            if m > 1:
                t += "^%d" % m

        if k > 0:
            t += "p"
            if k > 1:
                t += "^%d" % k
        return t

    ax.set_axis_off()
    for i in range(num_plates):
        x = dh*i + 0.03
        for k in range(num_plates):
            y = dh*(num_plates - 1 - k)
            ax.add_patch(Rectangle(xy=(x, y), width=w,  height=dh, alpha=0.5, edgecolor="blue"))
            if k <= i:
                tx = x + 0.5*w
                ty = y + 0.5*dh
                f = "2000"
                if i > 0:
                    f += r" \cdot "
                    f += term(i, k)
                ax.text(tx, ty, r"$ %s $" % f, ha="center", va="center")
                if i < num_plates - 1:
                    if False:
                        ax.arrow(tx+dh*0.2, ty, dh*0.5, 0, head_width=0.02, head_length=0.01, fc='k', ec='k', alpha=0.5)
                        ax.arrow(tx+dh*0.2, ty, dh*0.5, -dh*0.9, head_width=0.02, head_length=0.01, fc='k', ec='k', alpha=0.5,
                                arrowprops=dict(arrowstyle="-|>", connectionstyle="arc3"),
                                )
                    else:
                        ax.annotate("", xy=(tx+dh*0.8, ty), xycoords='data', xytext=(tx+dh*0.2, ty), textcoords='data',
                                arrowprops=dict(arrowstyle="-|>", connectionstyle="arc3"), alpha=0.5,
                                )
                        ax.annotate("", xy=(tx+dh*0.8, ty -dh*0.94), xycoords='data', xytext=(tx+dh*0.2, ty), textcoords='data',
                                arrowprops=dict(arrowstyle="-|>", connectionstyle="arc3"), alpha=0.5,
                                )

    fig.tight_layout()
    plt.show()

if __name__ == '__main__':
    # import seaborn
    # seaborn.set()
    # demo2(save=True)
    demo3()



"""
    Estimators.SdmEstimatorDebug.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

def components_plot_debug_impl(optimizer, params):
    print("components_plot_debug_impl")

    lrf_info = optimizer.objective_func(params, return_lrf_info=True)

    def plot_components(ax, x, y, C, ty):
        ax.plot(x, y, label="data")
        for k, cy in enumerate(C):
            ax.plot(x, cy, ":", label="component-%d" % k)
        ax.plot(x, ty, ":", color="red", label="model total")

    with plt.Dp():
        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(12,10))
        fig.suptitle("components_plot_debug_impl")
        ax1, ax2 = axes[0,0], axes[0,1]
        ax3, ax4 = axes[1,0], axes[1,1]
        plot_components(ax1, lrf_info.uv_x, lrf_info.uv_y, lrf_info.matrices[3], lrf_info.uv_ty)
        plot_components(ax2, lrf_info.x, lrf_info.y, lrf_info.matrices[1], lrf_info.xr_ty)
        plot_components(ax3, lrf_info.uv_x, lrf_info.uv_y, lrf_info.scaled_uv_cy_array, lrf_info.uv_ty)
        plot_components(ax4, lrf_info.x, lrf_info.y, lrf_info.scaled_xr_cy_array, lrf_info.xr_ty)
        fig.tight_layout()
        ret = plt.show()

    return ret


"""
    from:
        Recipe11.4.Adding Drag and Drop Reordering to a Tkinter Listbox
        https://flylib.com/books/en/2.9.1.230/1/
"""
import tkinter
class DDList(tkinter.Listbox):
    """ A tkinter listbox with drag'n'drop reordering of entries. """
    def __init__(self, master, **kw):
        kw['selectmode'] = tkinter.SINGLE
        tkinter.Listbox.__init__(self, master, kw)
        self.bind('<Button-1>', self.setCurrent)
        self.bind('<B1-Motion>', self.shiftSelection)
        self.curIndex = None
    def setCurrent(self, event):
        self.curIndex = self.nearest(event.y)
    def shiftSelection(self, event):
        i = self.nearest(event.y)
        if i < self.curIndex:
            x = self.get(i)
            self.delete(i)
            self.insert(i+1, x)
            self.curIndex = i
        elif i > self.curIndex:
            x = self.get(i)
            self.delete(i)
            self.insert(i-1, x)
            self.curIndex = i

if __name__ == '__main__':
    tk = tkinter.Tk( )
    length = 10
    dd = DDList(tk, height=length)
    dd.pack( )
    for i in range(length):
        dd.insert(tkinter.END, str(i))
    def show( ):
        ''' show the current ordering every 2 seconds '''
        for x in dd.get(0, tkinter.END):
            print(x),
        print()
    tk.after(2000, show)
    tk.after(2000, show)
    tk.mainloop( )



"""
    OurManim3D.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from mpl_toolkits.mplot3d import Axes3D 
# import matplotlib.pyplot as plt 
import matplotlib.animation as animation
import molass_legacy.KekLib.DebugPlot as plt
from OurManim import manim_init

def spike():
    manim_init()

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.view_init(0, 0)
    # ax.set_axis_off()
    ax.text(0.2, 0.2, 0.5, r"$ v $", ha="center", va="center", fontsize=100)
    ax.text(0.5, 0.5, 0.5, r"$ = $", ha="center", va="center", fontsize=100)
    ax.text(0.8, 0.8, 0.5, r"$ u $", ha="center", va="center", fontsize=100)
    plt.show()

class TextGroup3D:
    def __init__(self):
        pass



"""
    Models.Stochastic.MomentsCollation.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from molass_legacy.Models.Stochastic.MonoporeMoments import study_monopore_moments_impl

def plot_collation_state_impl(x, y, egh_cy_list, egh_moments_list, ax, moments_list):
    ax.plot(x, y, color="orange")

    for k, (cy, M, M_) in enumerate(zip(egh_cy_list, egh_moments_list, moments_list)):
        ax.plot(x, cy, ":")
        m, v = M[0:2]
        s = np.sqrt(v)
        ax.axvline(x=m, ls="-", color="blue")
        label = "EGH Moments" if k == 0 else None
        ax.axvspan(m-s, m+s, color="cyan", alpha=0.3, label=label)
        m_, v_ = M_[0:2]
        s_ = np.sqrt(v_)
        ax.axvline(x=m_, ls="-", color="red")
        label = "Stochastic Moments" if k == 0 else None
        ax.axvspan(m_-s_, m_+s_, color="pink", alpha=0.3, label=label)

    ax.legend(loc="upper left")

def moments_collation_illust_impl(lrf_src,  debug=False):
    print("moments_collation_illust_impl")

    rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = lrf_src.compute_rgs(debug=False)
    peaks = lrf_src.get_peaks()
    num_peaks = len(peaks)
    model = lrf_src.model
    egh_moments_list = lrf_src.get_egh_moments_list()
    x = lrf_src.xr_x
    y = lrf_src.xr_y
    egh_cy_list = []
    for params in peaks:
        cy = model(x, params)
        egh_cy_list.append(cy)
    egh_ty = np.sum(egh_cy_list, axis=0)

    def plot_collation_state(ax, moments_list):
        plot_collation_state_impl(x, y, egh_cy_list, egh_moments_list, ax, moments_list)

    init_moments_list  = study_monopore_moments_impl(lrf_src, return_init_moments=True)
    step1_moments_list = study_monopore_moments_impl(lrf_src, trust_all_rgs=True, use_basinhopping=False, return_step1_moments=True)
    step2_moments_list, unreliable_indeces = study_monopore_moments_impl(lrf_src, trust_max_num=2, return_step2_moments=True)

    with plt.Dp():
        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,5))
        fig.suptitle("Moments Collation Method Illustration on %s" % get_in_folder(), fontsize=20)
        ax1.set_title("Initial Guess", fontsize=16)
        plot_collation_state(ax1, init_moments_list)
        ax2.set_title("Optimize Simply with Local Optimization", fontsize=16)
        plot_collation_state(ax2, step1_moments_list)
        ax3.set_title("Optimize with Rg Coorection %s, Time Scaling, BH" % (unreliable_indeces), fontsize=16)
        plot_collation_state(ax3, step2_moments_list)
        fig.tight_layout()
        plt.show()


"""
    Theory.Prism.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

USE_PRISM = False
USE_PYOZ = False

def compute_rdf_prism():
    import pyPRISM
    sys = pyPRISM.System(['polymer'],kT=1.0)
    sys.domain = pyPRISM.Domain(dr=0.05,length=4096)
    sys.diameter['polymer'] = 1.0
    sys.density['polymer'] = 0.9
    sys.closure['polymer','polymer'] = pyPRISM.closure.PercusYevick()
    sys.potential['polymer','polymer'] = pyPRISM.potential.HardSphere()
    sys.omega['polymer','polymer'] = pyPRISM.omega.GaussianRing(sigma=sys.diameter['polymer','polymer'],length=2000)
    PRISM = sys.createPRISM()
    guess = np.zeros_like(sys.domain.r)
    result = PRISM.solve(guess)
    x = sys.domain.r
    y = pyPRISM.calculate.pair_correlation(PRISM)['polymer','polymer']
    return x, y

def compute_rdf_pyoz():
    import pyoz as oz
    lj_unary = oz.System()
    lj_unary.set_interaction(0, 0, oz.lennard_jones(lj_unary.r, eps=1.0, sig=1.0))
    r = lj_unary.r
    U_r = lj_unary.U_r
    g_r, c_r, e_r, h_k = lj_unary.solve(rhos=0.1, closure_name='hnc')
    return r, g_r[0,0]

def compute_sq(qv, r, g, rho=0.01):
    sy = np.zeros(len(qv))
    r2 = r**2
    scale = 4 * np.pi * rho
    for k, q in enumerate(qv):
        sy[k] = 1 + scale * np.sum((g - 1)*np.sinc(q*r)*r2)
    return sy

def compute_rdf(qv, sy, rv, rho=0.01):
    gy = np.zeros(len(rv))
    scale = 1/(2*np.pi**2*rho)
    q2 = qv**2
    for i, r in enumerate(rv):
        gy[i] = 1 + scale * np.sum((sy - 1)*np.sinc(qv*r)*q2)
    return gy

def demo(sp, sd, pno=0):
    from bisect import bisect_right
    from Theory.Rg import compute_corrected_Rg
    from SvdDenoise import get_denoised_data
    from Theory.SolidSphere import phi, get_boundary_params_simple
    from Theory.SynthesizedLRF import synthesized_lrf_spike
    from .DjKinning1984 import P0, S0

    M, E, qv, ecurve = sd.get_xr_data_separate_ly()

    range_ = ecurve.get_ranges_by_ratio(0.5)[pno]
    f = range_[0]
    p = range_[1]
    t = range_[2]

    pt = M[:,p]

    x = ecurve.x
    y = ecurve.y

    eslice = slice(f,t)
    M0 = M[:,eslice]
    Rg, gf, gt, _ = compute_corrected_Rg(sd, ecurve, pno, qv, M0, E[:,eslice])
    R = np.sqrt(5/3)*Rg

    if True:
        i = np.argmax(pt)
    else:
        # q*5.4*R/2 == np.pi/2
        q0 = np.pi/(5.4*R)
        print("q0=", q0)
        i = bisect_right(qv, q0)

    p0 = P0(qv, R)
    s0 = S0(qv, R/2*5.4, K=1)

    norm_list = []

    c_ = y[eslice]

    M2 = get_denoised_data(M0, rank=2)
    C2 = np.array([c_, c_**2])
    P2 = M2 @ np.linalg.pinv(C2)

    b1, b2, k = get_boundary_params_simple(Rg)
    P_ = synthesized_lrf_spike(qv, M0, c_, M2, P2, boundary=b1, k=k)
    y1 = P_[:,0]
    y1n = y1/y1[0]
    s1 = 1 + P_[:,1]/P_[:,0]

    rho = 0.005

    if USE_PRISM:
        gxp, gyp = compute_rdf_prism()
        s2r = compute_sq(qv, gxp, gyp)

    gx = np.linspace(0, 500, 1000)
    g0 = compute_rdf(qv, s0, gx, rho=rho)
    g1 = compute_rdf(qv, s1, gx, rho=rho)

    if USE_PYOZ:
        gx_, g_ = compute_rdf_pyoz()
        s_r = compute_sq(qv, gx_, g_, rho=rho)

    s0r = compute_sq(qv, gx, g0, rho=rho)
    s1r = compute_sq(qv, gx, g1, rho=rho)

    fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(21,11))
    ax1, ax2, ax3 = axes[0,:]
    ax4, ax5, ax6 = axes[1,:]

    fig.suptitle("S(q) recomputation from g(r)", fontsize=20)
    ax1.set_title("S(q)", fontsize=16)
    ax2.set_title("P(q)*S(q)", fontsize=16)
    ax3.set_title("g(r)", fontsize=16)
    ax4.set_title("S(q) recomputed from g(r)", fontsize=16)

    ax2.set_yscale('log')

    label0 = "Hard Sphere"
    label1 = "GI (Synthesized LRF)"

    ax1.plot(qv, s0, label=label0)
    ax1.plot(qv, s1, label=label1)

    y0 = p0*s0
    ax2.plot(qv, y0, label=label0)
    y1n = y1/y1[0]
    ax2.plot(qv, y1n, label=label1)

    ax3.plot(gx, g0, label=label0)
    ax3.plot(gx, g1, label=label1)

    ax4.plot(qv, s0r, label=label0)
    ax4.plot(qv, s1r, label=label1)

    if USE_PRISM:
        ax3.plot(gxp, gyp)
        ax4.plot(qv, s2r)

    if USE_PYOZ:
        ax3.plot(gx_, g_)
        ax4.plot(qv, s_r)

    for ax in [ax1, ax2, ax3, ax4]:
        ax.legend(fontsize=16)

    fig.tight_layout()
    fig.subplots_adjust(top=0.9)
    plt.show()



"""
    UvBaseSolver.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from scipy.optimize import basinhopping
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt
from MatrixData import simple_plot_3d
from SvdDenoise import get_denoised_data
from molass_legacy.Peaks.EghSupples import egh, d_egh
from .Sigmoid import ex_sigmoid

class UvBaseSolver:
    def __init__(self, wv, A, info, debug=False):
        self.info = info
        self.wv = wv
        self.A = A
        self.indeces = [bisect_right(wv, w) for w in [280, 340]]

        if debug:
            egh_param_list, sigmoid_param_list, d_egh_param_list = info
            i1, i2 = self.indeces
            with plt.Dp():
                fig = plt.figure(figsize=(18,5))
                fig.suptitle("UvBaseSolver.__init__ debug")
                ax1 = fig.add_subplot(131, projection="3d")
                ax2 = fig.add_subplot(132)
                ax3 = fig.add_subplot(133)

                simple_plot_3d(ax1, A, x=wv)

                x = np.arange(A.shape[1])

                for ax, i in [(ax2, i1), (ax3, i2)]:
                    ax.plot(x, A[i,:])

                for params in egh_param_list:
                    ax2.plot(x, egh(x, *params))

                for params in sigmoid_param_list:
                    ys = ex_sigmoid(x, *params)
                    ax3.plot(x, ys)

                for k, params in enumerate(egh_param_list):
                    params[0] = d_egh_param_list[k]
                    yd = d_egh(x, *params)
                    ax3.plot(x, ys + yd)

                fig.tight_layout()
                plt.show()

    def solve(self, optimize=False, debug=False):
        U = get_denoised_data(self.A, rank=2)

        egh_param_list, sigmoid_param_list, d_egh_param_list = self.info

        egh_param_array = np.array(egh_param_list)
        sigmoid_param_array = np.array(sigmoid_param_list)
        d_egh_param_array = np.array(d_egh_param_list)

        np1 = np.product(egh_param_array.shape)
        np2 = np1 + np.product(sigmoid_param_array.shape)

        x = np.arange(U.shape[1])
        egh_param_list_ = [prm.copy()for prm in egh_param_list]

        def compute_PQ(p):
            egh_params = p[0:np1].reshape(egh_param_array.shape)
            sigmoid_params = p[np1:np2].reshape(sigmoid_param_array.shape)
            d_egh_params = p[np2:]

            c = np.zeros(len(x))
            for prms in egh_params:
                c += egh(x, *prms)

            g = np.zeros(len(x))
            for prms in sigmoid_params:
                g += ex_sigmoid(x, *prms)

            for k, prms in enumerate(egh_param_list_):
                prms[0] = d_egh_params[k]
                g += d_egh(x, *prms)

            Q = np.array([c, g])
            P = U @ np.linalg.pinv(Q)
            return P, Q

        p0 = np.concatenate([array.flatten() for array in [egh_param_array, sigmoid_param_array, d_egh_param_array]])

        P0, Q0 = compute_PQ(p0)
        U0 = P0 @ Q0
        U00 = P0[:,[0]] @ Q0[[0],:]
        U01 = P0[:,[1]] @ Q0[[1],:]

        if debug:
            from matplotlib.gridspec import GridSpec
            from DataUtils import get_in_folder
            wv = self.wv
            A = self.A
            gs = GridSpec(9,3)
            # fig, (ax1, ax2, ax3, ax4) = plt.subplots(ncols=4, figsize=(24,5), subplot_kw=dict(projection="3d"))
            with plt.Dp():
                fig = plt.figure(figsize=(15,12))
                ax1, ax2, ax3 = [fig.add_subplot(gs[0:3,k], projection="3d")  for k in range(3)]
                ax3u = fig.add_subplot(gs[3:6,2], projection="3d")
                ax3d = fig.add_subplot(gs[6:9,2], projection="3d")
                fig.suptitle("Trial Result of the New Correction Method based on Rank2 LRF for %s" % get_in_folder(), fontsize=20)
                ax1.set_title("Input Data", fontsize=16)
                ax2.set_title("Rank2 denoised", fontsize=16)
                ax3.set_title("Reconstructed with Rank2 LRF", fontsize=16)
                ax3u.set_title("Result (First Component only)", fontsize=16)

                simple_plot_3d(ax1, self.A)
                simple_plot_3d(ax2, U)
                simple_plot_3d(ax3, U0)
                simple_plot_3d(ax3u, U00)
                zmin, zmax = ax3u.get_zlim()
                ax3d.set_zlim( zmin, zmax)
                simple_plot_3d(ax3d, U01)
                for ax in [ax1, ax2, ax3, ax3u, ax3d]:
                    ax.view_init(15, 30)
                fig.tight_layout()
                plt.show()

        self.U00 = U00
        self.U01 = U01
        if not optimize:
            self.U00_ = None
            self.U01_ = None
            return

        i1, i2 = self.indeces

        def objective_func(p):
            P, Q = compute_PQ(p)
            return np.linalg.norm(P@Q - U)

        result = minimize(objective_func, p0)
        # result = basinhopping(objective_func, p0)

        p0_ = result.x
        P0_, Q0_ = compute_PQ(p0_)
        U0_ = P0_ @ Q0_
        U00_ = P0_[:,[0]] @ Q0_[[0],:]
        U01_ = P0_[:,[1]] @ Q0_[[1],:]

        print("objective values=", [objective_func(p) for p in [p0, p0_]])

        self.U00_ = U00_
        self.U01_ = U01_

    def get_baseline(self, debug=False):
        A = self.A
        i1, i2 = self.indeces

        if debug:
            from matplotlib.gridspec import GridSpec
            from DataUtils import get_in_folder
            wv = self.wv
            x = np.arange(A.shape[1])

            def plot_resuls(ax0, ax1, ax2, ax3, ax4, U0, U1):
                simple_plot_3d(ax0, A, x=wv)
                zmin, zmax = ax0.get_zlim()
                for ax in [ax1, ax2]:
                    ax.set_zlim(zmin, zmax)
                simple_plot_3d(ax1, U0, x=wv)
                simple_plot_3d(ax2, U1, x=wv)

                a1 = A[i1,:]
                b1 = U1[i1,:]
                ax3.plot(x, a1, label="input data")
                ax3.plot(x, a1 - b1, label="corrected")
                # ax3.plot(x, U0[i1,:], label="egh model")
                ax3.plot(x, U1[i1,:], color="red", label="LRF baseline")
                ax3.legend()

                a2 = A[i2,:]
                b2 = U1[i2,:]
                ax4.plot(x, a2, label="input data")
                ax4.plot(x, a2 - b2, label="corrected")
                ax4.plot(x, U1[i2,:], color="red", label="LRF baseline")
                ax4.legend()

            with plt.Dp():
                # fig, axes = plt.subplots(ncols=2, nrows=2, figsize=(10,8))
                U_list = [(self.U00, self.U01)]
                if self.U00_ is not None:
                    U_list.append((self.U00_, self.U01_))
                nrows = len(U_list)
                gs = GridSpec(nrows,5)
                fig = plt.figure(figsize=(23,4*nrows))
                for k, upair in enumerate(U_list):
                    ax0 = fig.add_subplot(gs[k,0], projection="3d")
                    ax1 = fig.add_subplot(gs[k,1], projection="3d")
                    ax2 = fig.add_subplot(gs[k,2], projection="3d")
                    ax3 = fig.add_subplot(gs[k,3])
                    ax4 = fig.add_subplot(gs[k,4])
                    plot_resuls(ax0, ax1, ax2, ax3, ax4, *upair)
                fig.tight_layout()
                plt.show()

        return self.U01[i1,:].copy()

def get_lrf_elution_base(sd, pre_recog, debug=False):
    from molass_legacy._MOLASS.SerialSettings import get_setting

    U, _, wv, ecurve = sd.get_uv_data_separate_ly()
    info = pre_recog.flowchange.get_solver_info(debug=debug)
    ubs = UvBaseSolver(wv, U, info)
    ubs.solve(debug=debug)
    base = ubs.get_baseline(debug=False)
    uv_restrict_list = get_setting('uv_restrict_list')
    if uv_restrict_list is None:
        eslice = slice(None, None)
    else:
        elution_restrict = uv_restrict_list[0]
        if elution_restrict is None:
            eslice = slice(None, None)
        else:
            eslice = slice(elution_restrict.start, elution_restrict.stop)
    return base[eslice].copy()



"""
    EnvironCheck.py

    Copyright (c) 2018-2024, Masatsuyo Takahashi, KEK-PF
"""

def executables_check(parent):
    try:
        import numba
        numba_is_available = True
    except:
        numba_is_available = False

    if numba_is_available:
        ret = True
    else:
        from molass_legacy._MOLASS.SerialSettings import get_setting
        suppress_numba_warning = get_setting('suppress_numba_warning')
        if suppress_numba_warning:
            ret = True
        else:
            import OurMessageBox as MessageBox
            yn = MessageBox.askyesno("No numba warning",
                '"numba" does not seem to be available.\n'
                "It would be desirable if you could take some time to update.\n"
                "Or, extrapolation with smoothing will be significantly slow.\n"
                "Would you like to proceed anyway?",
                parent=parent)
            ret = yn
    return ret



# coding: utf-8
"""
    RgProcess.RgAnalysis.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
try:
    import molass_legacy.KekLib.CustomMessageBox as MessageBox
except:
    import OurMessageBox as MessageBox
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.widgets import Button
from .RgCurve import RgCurve

class RgAnalysisDialog(Dialog):
    def __init__(self, parent, si):
        parent.report_callback_exception = self.report_callback_exception
        self.parent = parent
        self.si = si
        print('X=', si.X.shape)
        print('curve_xy=', si.curve_xy.shape)
        self.xy = si.curve_xy
        self.q = si.X[0,:,0]
        self.q2 = self.q**2
        self.D = si.X[:,:,1].T
        self.E = si.X[:,:,2].T

        self.in_folder = str(si.in_folder)
        Dialog.__init__(self, parent, "Guinier Analysis", visible=False)

    def report_callback_exception(self, exc, val, tb):
        # This method is to override the Tk method to be able
        # to report to the spawned console in windows application mode.
        from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker

        etb = ExceptionTracebacker()
        msg = ( 'Overridden report_callback_exception: ' + str( etb ) )
        MessageBox.showerror( "Tkinter Error", msg, parent=self.parent )
        self.parent,quit()

    def show(self):
        self._show()

    def body(self, body_frame):
        dummy_label = Tk.Label(body_frame, text=self.in_folder)
        dummy_label.pack()

        self.cframe = Tk.Frame(body_frame)
        self.cframe.pack()

        fig, axes = plt.subplots(ncols=3, figsize=(21,7))

        self.fig = fig
        self.axes = axes

        self.mpl_canvas = FigureCanvasTkAgg(fig, self.cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        self.j = 150

        self.L_button = Button(plt.axes([0.04, 0.5, 0.04, 0.04]), '$ < $')
        self.L_button.on_clicked(lambda *args: self.move_step(-1))
        self.R_button = Button(plt.axes([0.27, 0.5, 0.04, 0.04]), '$ > $')
        self.R_button.on_clicked(lambda *args: self.move_step(+1))

        fig.tight_layout()
        self.reraw()

    def reraw(self):
        self.draw_ecurve(self.axes[0])
        self.draw_scurve(self.axes[1:])
        self.mpl_canvas.draw()

    def draw_ecurve(self, ax):
        x, y = self.xy
        ax.plot(x, y)
        j = self.j
        ax.plot(x[j], y[j], 'o', color='yellow')

    def draw_scurve(self, axes):
        j = self.j
        x = self.q2
        y = np.log(self.D[:,j])
        for ax in axes:
            ax.plot(x, y)

    def move_step(self, i):
        self.j = max(0, min(self.D.shape[1], self.j+i))
        print("move_step to", self.j)
        for ax in self.axes:
            ax.cla()
        self.reraw()



# coding: utf-8
"""

    SerialTestUtils.py

    Copyright (c) 2018-2021, SAXS Team, KEK-PF

"""
import numpy                as np
from molass_legacy.KekLib.BasicUtils             import clear_dirs_with_retry, Struct
from molass_legacy.KekLib.ExceptionTracebacker   import ExceptionTracebacker
from ChangeableLogger       import Logger
from molass_legacy._MOLASS.SerialSettings         import set_setting
from SerialDataUtils        import get_mtd_filename
from molass_legacy.SerialAnalyzer.AbnormalityCheck       import bubble_check
from InputSmootherAveraging import ConcentrationSmootherAveraging
from molass_legacy.KekLib.OurTkinter             import Tk
from molass_legacy.KekLib.DebugPlot              import set_plot_env
from molass_legacy.KekLib.TkUtils                import adjusted_geometry, get_tk_root

def create_our_logger(temp_dir):
    # clear_dirs_with_retry( [ temp_dir ] )
    logger = Logger( temp_dir + '/test.log' )
    return logger

def standard_setup(temp_dir):
    clear_dirs_with_retry( [temp_dir] )
    logger = create_our_logger( temp_dir )
    set_setting( 'temp_folder', temp_dir )
    set_setting( 'analysis_folder', temp_dir )
    return logger

def debug_plot_runner(run_func):
    root = Tk.Tk()
    root.geometry( adjusted_geometry( root.geometry() ) )
    root.update()
    root.withdraw()
    set_plot_env(root)

    def run_closure():
        run_func()
        root.quit()

    root.after(0, run_closure)
    root.mainloop()
    root.destroy()

def get_serialdata_simply(in_folder):
    from SerialDataLoader import SerialDataLoader
    sdl = SerialDataLoader()
    sdl.load_from_folders(in_folder, in_folder)
    sd = sdl.get_data_object()
    return sd

def prepare_serialdata_env( in_folder, uv_folder=None, conc_factor=5, mapper_hook=None, enable_auto_helper=True,
                            xray_only=False, mtd_conc=False, adjust=False, root=None, logger=None, ret_pre_recog=False ):
    from SerialDataLoader import SerialDataLoader
    set_setting( 'in_folder', in_folder )   # for xray 3d plot
    if enable_auto_helper:
        set_setting( 'enable_auto_helper', 1 )
    else:
        set_setting( 'enable_auto_helper', 0 )

    if uv_folder is None:
        uv_folder = in_folder

    sdl = SerialDataLoader()
    if xray_only or mtd_conc:
        if mtd_conc:
            set_setting( 'use_mtd_conc', 1 )
            set_setting( 'use_xray_conc', 0 )
            filename = get_mtd_filename( in_folder )
            set_setting( 'mtd_file_path', filename )
        else:
            set_setting( 'use_mtd_conc', 0 )
            set_setting( 'use_xray_conc', 1 )
        sdl.load_xray_data_only( in_folder )
    else:
        sdl.load_from_folders( in_folder, uv_folder )
    sd  = sdl.get_data_object()

    exclude = bubble_check( sd )
    if len(exclude) > 0:
        sd.exclude_intensities( exclude )
        sdl.memorize_exclusion( sd )

    if ret_pre_recog or adjust:
        if sd.is_serial():
            from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
            try:
                pre_recog = PreliminaryRecognition(sd)
                sd.conc_factor = conc_factor
            except:
                etb = ExceptionTracebacker()
                print(etb.last_lines())
                pre_recog = None

    if adjust:
        # this will do the baseline correction
        sd = pre_recog.get_analysis_copy()
        mapper = get_mapper_for_the_test(root, sd, adjust=adjust, logger=logger)

    if ret_pre_recog:
        return sd, pre_recog
    else:
        return sd

global_info = None

def get_mapper_for_the_test(root, sd_, adjust=False, default_opt_params=False, opt_methods=None, logger=None):
    from molass_legacy.Mapping.MapperConstructor import create_mapper

    if logger is not None:
        logger.info('get_mapper_for_the_test: adjust=' + str(adjust))

    if default_opt_params:
        opt_params = None
    else:
        from molass_legacy.Mapping.MappingParams import MappingParams

        opt_params_default = MappingParams(
            ( 'uv_baseline_opt',            1 ),
            ( 'uv_baseline_type',           1 ),
            ( 'uv_baseline_adjust',         0 ),
            ( 'uv_baseline_with_bpa',       0 ),
            ( 'xray_baseline_opt',          0 ),
            ( 'xray_baseline_type',         0 ),
            ( 'xray_baseline_adjust',       0 ),
            ( 'xray_baseline_with_bpa',     0 ),
            ( 'dev_allow_ratio',            0.5 ),
            )

        opt_params_adjust = MappingParams(
            ( 'uv_baseline_opt',            1 ),
            ( 'uv_baseline_type',           1 ),
            ( 'uv_baseline_adjust',         1 ),
            ( 'uv_baseline_with_bpa',       0 ),
            ( 'xray_baseline_opt',          1 ),
            ( 'xray_baseline_type',         0 ),
            ( 'xray_baseline_adjust',       1 ),
            ( 'xray_baseline_with_bpa',     1 ),
            ( 'dev_allow_ratio',            0.5 ),
            )

        if adjust:
            opt_params = opt_params_adjust
        else:
            opt_params = opt_params_default

    mapper = create_mapper( root, sd_, opt_params=opt_params, opt_methods=opt_methods, logger=logger )
    return mapper

def get_tttt_data_objects( in_folder, uv_folder=None, xray_only=False, adjust=False,
                            exec_copy=False, analysis_copy=False,
                            default_opt_params=False, root=None, logger=None,
                            opt_methods=None, mtd_conc=False ):
    global global_info

    sd_orig = prepare_serialdata_env( in_folder, uv_folder, xray_only=xray_only, mtd_conc=mtd_conc )
    # sd.absorbance.solve_bottomplane_LPM()



    if root is None:
        root = Tk.Tk()
        root.geometry( adjusted_geometry( root.geometry() ) )
        root.withdraw()

    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    pre_recog = PreliminaryRecognition(sd_orig)

    if analysis_copy:
        # sd_ = pre_recog.get_analysis_copy()
        sd_ = sd_orig._get_analysis_copy_impl(pre_recog)
    else:
        sd_ = sd_orig

    mapper = get_mapper_for_the_test(root, sd_, adjust=adjust, default_opt_params=default_opt_params, opt_methods=opt_methods, logger=logger)

    conc_factor = 5
    if exec_copy:
        sd = sd_.get_exec_copy( mapper, conc_factor )
        sd.absorbance = sd_orig.absorbance    # exec_copy does not have absorbance
    else:
        sd = sd_
        sd.set_mc_vector(mapper, conc_factor)

    global_info = Struct( parent=root, sd=sd, mapper=mapper )

    return root, sd, mapper

def get_global_info():
    global global_info
    if global_info is None:
        raise RuntimeError( "call get_tttt_data_objects before getting global info" )
    return global_info

def set_setting_for_baseline_correction( opt_params ):
    set_setting( 'uv_baseline_opt',         opt_params[ 'uv_baseline_opt' ] )
    set_setting( 'uv_baseline_const_opt',   opt_params[ 'uv_baseline_const_opt' ] )
    set_setting( 'uv_baseline_adjust',      opt_params[ 'uv_baseline_adjust' ] )
    set_setting( 'xray_baseline_opt',       opt_params[ 'xray_baseline_opt' ] )
    set_setting( 'xray_baseline_const_opt', opt_params[ 'xray_baseline_const_opt' ] )
    set_setting( 'xray_baseline_adjust',    opt_params[ 'xray_baseline_adjust' ] )

class MockExecuter:
    def __init__( self, sd ):
        from molass_legacy.Mapping.ElutionMapper import ElutionMapper
        mapper  = ElutionMapper( sd )
        mapper.optimize()
        conc_factor = 5
        self.mapped_info    = mapper.get_mapped_info()
        sd.set_mc_vector( mapper, conc_factor )

        # Concetration Smoothing
        cvector_smoother = ConcentrationSmootherAveraging( sd.mc_vector, 5 )
        indeces = np.arange( 0, sd.intensity_array.shape[0] )
        c_vector = cvector_smoother( indeces )
        c_vector[ np.abs( c_vector ) < 1e-5 ] = 1e-5

        self.c_vector       = c_vector
        # self.quality_array  = 

def get_uv_elution_vector( sd, correct_base=True, use_LB=False, plot=False):
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from molass_legacy.UV.Absorbance import Absorbance

    absorbance = Absorbance( sd.lvector, sd.conc_array, sd.xray_curve, col_header=sd.col_header )
    a_vector    = absorbance.a_vector

    if correct_base:
        try:
            pre_recog = PreliminaryRecognition(sd)
            j_min, j_max = pre_recog.get_flow_changes()
            if use_LB:
                absorbance.solve_bottomplane_LB(j_min, j_max)
            else:
                absorbance.solve_bottomplane_LPM(j_min, j_max)
            base    = absorbance.get_standard_elution_base()
        except:
            # when mtd_conc==True
            base = 0
    else:
        base = 0

        if plot:
            import molass_legacy.KekLib.DebugPlot as plt
            plt.plot( a_vector )
            if type(base) != int:
                plt.plot( base, color='red' )
            plt.show()

    return a_vector - base



"""
    Peaks.PeakDevel.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
from importlib import reload
import molass_legacy.KekLib.DebugPlot as plt

def devel_test_impl(self):
    print("devel_test_impl")

    def exec_embed_cushion():
        import molass_legacy.Tools.EmbedCushion
        reload(molass_legacy.Tools.EmbedCushion)
        from molass_legacy.Tools.EmbedCushion import embed_cushion
        embed_cushion(self)

    def cpd_spike():
        import molass_legacy.GuinierTools.CpdDecompIndirect
        reload(molass_legacy.GuinierTools.CpdDecompIndirect)
        from molass_legacy.GuinierTools.CpdDecompIndirect import cpd_spike_impl
        cpd_spike_impl(self)

    def estimator_test():
        import molass_legacy.Estimators.TestTools
        reload(molass_legacy.Estimators.TestTools)
        from molass_legacy.Estimators.TestTools import estimator_test_impl
        estimator_test_impl(self)

    def show_restart_patcher():
        import molass_legacy.Optimizer.RestartPatcher
        reload(molass_legacy.Optimizer.RestartPatcher)
        from molass_legacy.Optimizer.RestartPatcher import patch_and_restart_from_editor
        patch_and_restart_from_editor(self)

    def test_fixedbaseline_optimizer():
        import molass_legacy.Optimizer.FixedBaselineOptimizer
        reload(molass_legacy.Optimizer.FixedBaselineOptimizer)
        from molass_legacy.Optimizer.FixedBaselineOptimizer import test_optimizer
        test_optimizer(self)

    def debug_objective_function():
        import molass_legacy.Optimizer.SimpleDebugUtils
        reload(molass_legacy.Optimizer.SimpleDebugUtils)
        from molass_legacy.Optimizer.SimpleDebugUtils import debug_optimizer
        debug_optimizer(self.fullopt, self.init_params)

    def test_estimate_uvbaseline():
        import molass_legacy.Optimizer.UvBaselineEstimator
        reload(molass_legacy.Optimizer.UvBaselineEstimator)
        from molass_legacy.Optimizer.UvBaselineEstimator import test_estimate_uvbaseline_impl
        test_estimate_uvbaseline_impl(self.fullopt, self.init_params)

    def test_func_loader():
        import logging
        import molass_legacy.Optimizer.FuncImporter
        reload(molass_legacy.Optimizer.FuncImporter)
        from molass_legacy.Optimizer.FuncImporter import get_objective_function_info
        logger = logging.getLogger(__name__)
        func_info = get_objective_function_info(logger, default_func_code='G0346')
        print("Objective Functions:", func_info.func_dict)

    extra_button_specs = [
        ("Embed Cushion", exec_embed_cushion),
        ("CPD Spike", cpd_spike),
        ("Estimator Test", estimator_test),
        ("Restart Patcher", show_restart_patcher),
        ("Fixed Baseline Optimizer Test", test_fixedbaseline_optimizer),
        ("Objective Function Debug", debug_objective_function),
        ("Estimate UV Baseline", test_estimate_uvbaseline),
        ("Function Loader Test", test_func_loader),
    ]

    with plt.Dp(button_spec=["OK", "Cancel"], extra_button_specs=extra_button_specs):
        fig, ax = plt.subplots()
        plt.show()



"""
    Models/RateTheory/RobustEDM_Demo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
def demo(in_folder, sd):
    from importlib import reload
    import Models.RateTheory.RobustEDM
    reload(Models.RateTheory.RobustEDM)
    from molass_legacy.Models.RateTheory.RobustEDM import guess_multiple

    ecurve = sd.get_xray_curve()
    x = ecurve.x.copy()
    y = ecurve.y

    guess_multiple(x, y, debug=True)



"""
    Characteristic.py

    Copyright (c) 2023-2025, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.SecTheory.SecPDF import compute_standard_wCD

class CfSpace:
    def __init__(self, N=1024, use_np_fft=True):
        self.use_np_fft = use_np_fft
        if use_np_fft:
            self.w = np.fft.fftfreq(N)
        else:
            w, C, D = compute_standard_wCD(N)
            self.w = w
            self.C = C
            self.D = D

    def get_w(self):
        return self.w

    def compute_cf(self, x, y):
        self.area = np.sum(y)
        y_ = y/self.area
        if self.use_np_fft:
            cft = np.fft.fft(y_, n=len(self.w))
        else:
            cft = []
            for w_ in self.w:
                cft.append(np.sum(np.exp(1j*w_*x)*y_))
            cft = np.array(cft)
        return cft



# coding: utf-8
"""
    ファイル名：   OurImageIO.py

    処理内容：

        画像ファイルの種類による差異を吸収するための、画像入出力クラス

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF
"""

import re
from MinimalTiff        import MinimalTiff
import fabio

extension_dict = {
    'tiff'  :   'tif',
    }

extension_re = re.compile( '\.(\w+)$' )

class Image:
    def __init__( self, path ):
        m = extension_re.search( path )
        if m:
            ext_ = m.group( 1 ).lower()
        else:
            assert( False )

        if len( ext_ ) != 3:
            ext_  = extension_dict[ ext_ ]

        self.__ext = ext_

        if ext_ == 'tif':
            self.__image    = MinimalTiff( path )
            self.__header   = self.__image.header
            self.__data     = self.__image.data
        elif ext_ == 'cbf':
            self.__image    = fabio.open( path )
            self.__data     = self.__image.data
        else:
            assert( False )

    def get_header( self )  : return self.__header
    def set_header( self )  : assert( False )
    header  = property( get_header, set_header )

    def get_data( self )    : return self.__data
    def set_data( self, data, force=False ):
        self.__data = data
        self.__image.data = data
    data    = property( get_data, set_data )

    def get_ext( self )     : return self.__ext
    def set_ext( self )     : assert( False )
    ext     = property( get_ext, set_ext )

    def save( self, path ):
        if self.__ext == 'tif':
            self.__image.save( path )
        elif self.__ext  == 'cbf':
            self.__image.save( path )
        else:
            assert( False )



"""
    DevTrial.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
from importlib import reload
import molass_legacy.KekLib.DebugPlot as plt

def dev_trial_entry(caller):

   corrected_sd = None
   comparison_result = None

   def prepare_data():
      import Trimming.MomentTrimming
      reload(Trimming.MomentTrimming)
      from molass_legacy.Trimming.MomentTrimming import set_moment_trimming_info
      import Batch.LiteBatch
      reload(Batch.LiteBatch)
      from molass_legacy.Batch.LiteBatch import create_corrected_sd
      nonlocal corrected_sd

      set_moment_trimming_info(caller.serial_data)
      corrected_sd = create_corrected_sd(caller.serial_data)
      ax.cla()
      ax.text(0.5, 0.5, "Corrected Data Prepared", ha="center", va="center", fontsize=30, color="gray")
      fig.canvas.draw()

   def spike_test():
      import Models.MomentFitting
      reload(Models.MomentFitting)
      from molass_legacy.Models.MomentFitting import spike
      spike(caller)
 
   def bridge_test():
      import Alsaker.Bridge
      reload(Alsaker.Bridge)
      from molass_legacy.Alsaker.Bridge import bridge_test_impl
      bridge_test_impl(caller)

   def compare_bridge():
      import Alsaker.Compare
      reload(Alsaker.Compare)
      from molass_legacy.Alsaker.Compare import compare_bridge_impl
      nonlocal comparison_result
      comparison_result = compare_bridge_impl(caller, corrected_sd=corrected_sd)

   def plot_comparison_result():
      import Alsaker.ComparisonPlot
      reload(Alsaker.ComparisonPlot)
      from molass_legacy.Alsaker.ComparisonPlot import plot_comparison_result_impl
      plot_comparison_result_impl(corrected_sd, comparison_result)

   def v2_baseline_inspect():
      import Baseline.Inspect
      reload(Baseline.Inspect)
      from molass_legacy.Baseline.Inspect import baseline_inspect_impl
      baseline_inspect_impl(caller)

   def bo_baseline_trial():
      import Baseline.BoBaseline
      reload(Baseline.BoBaseline)
      from molass_legacy.Baseline.BoBaseline import bo_baseline_trial_impl
      bo_baseline_trial_impl(caller)

   def debug_modeled_peaks():
      import QuickAnalysis.PeakUtils
      reload(QuickAnalysis.PeakUtils)
      from molass_legacy.QuickAnalysis.PeakUtils import demo_modeled_peaks_dialog
      in_folder = caller.in_folder.get()
      demo_modeled_peaks_dialog(caller, in_folder)

   def opt_view_range():
      import Trimming.OptViewRange
      reload(Trimming.OptViewRange)
      from molass_legacy.Trimming.OptViewRange import get_opt_view_range
      
      get_opt_view_range(caller)

   extra_button_specs = [
      ("Prepare Dara", prepare_data),
      ("MomentFitting", spike_test),
      ("Alsaker Bridge Test", bridge_test),
      ("Compare Bridge to SimpleGuinier", compare_bridge),
      ("Plot Comparison Result", plot_comparison_result),
      ("V2 Baseline", v2_baseline_inspect),
      ("BO Baseline", bo_baseline_trial),
      ("Debug Modeled Peaks", debug_modeled_peaks),
      ("Opt View Range", opt_view_range),
      ]

   with plt.Dp(button_spec=["Close"],
               extra_button_specs=extra_button_specs):
      fig, ax = plt.subplots()
      plt.show()


# coding: utf-8
"""
    SecTheory.Dondi1986.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib.gridspec import GridSpec
import molass_legacy.KekLib.DebugPlot as plt

def demo():
    lam = 40
    size = 1000
    Np = np.random.poisson(lam, size)

    if False:
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.plot(Np)
            plt.show()

    fig, ax = plt.subplots()
    scale = 1.0
    traj_list = []
    for n in Np:
        y = np.linspace(0, 1, n)
        Tm = np.random.exponential(scale, n)
        Ts = np.random.exponential(scale, n)
        tx = [0.0]
        ty = [0.0]
        t = 0
        my = 0
        for k, (tm, ts) in enumerate(zip(Tm, Ts)):
            t += tm
            my += tm
            tx.append(t)
            ty.append(my)
            if k < n-1:
                t += ts
                tx.append(t)
                ty.append(my)
        tx = np.array(tx)
        ty = np.array(ty)/my
        traj_list.append((tx, ty))

    with plt.Dp():
        fig = plt.figure(figsize=(10,10))
        fig.suptitle("Stochastic Theory Illustration with size=%d" % size, fontsize=20)

        gs = GridSpec(5,1)

        ax1 = fig.add_subplot(gs[0,:])
        ax2 = fig.add_subplot(gs[1:,:])

        ax2.set_xlabel("Time", fontsize=16)
        ax2.set_ylabel("Length (Displacement)", fontsize=16)

        tr = [tx[-1] for tx, ty in traj_list]
        ax1.hist(tr)

        for tx, ty in traj_list:
            ax2.plot(tx, ty)

        xmin, xmax = ax2.get_xlim()
        ax.set_xlim(xmin, xmax)

        fig.tight_layout()
        plt.show()



"""
    UV.WYF_RatiosSurvey.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
import sys
import numpy as np

def get_wyf_ratios(in_folder, *args, **kwargv):
    from molass_legacy._MOLASS.SerialSettings import clear_temporary_settings, clear_v2_temporary_settings
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Batch.LiteBatch import LiteBatch
    from molass_legacy.UV.WYF_Ratios import compute_ratios_from_Puv
    from molass_legacy.SerialAnalyzer.DataUtils import cut_upper_folders

    global fh

    print(in_folder)
    folder_name = cut_upper_folders(in_folder)
    if False:
    # if folder_name < "20170506":
    # if folder_name.find("OA001") < 0:
        print("skip")
        return True, None

    try:
        clear_temporary_settings()
        clear_v2_temporary_settings()
        sp = StandardProcedure()
        sd = sp.load_old_way(in_folder)
        lb = LiteBatch()
        lb.prepare(sd)
        init_params = lb.get_init_estimate()

        lb.optimizer.prepare_for_optimization(init_params, minimally=True)
        lrf_info = lb.optimizer.objective_func(init_params, return_lrf_info=True)
        proportions = lrf_info.get_xr_proportions()
        print("proportions=", proportions)
        Pxr, Cxr, Puv, Cuv, mapped_UvD = lrf_info.matrices
        ret_list = compute_ratios_from_Puv(lb.optimizer.wvector, Puv[:,:-1], proportions=proportions)
    except:
        ret_list = [(np.nan, np.nan)]

    for k, ratios in enumerate(ret_list):
        peak_name = "%s-%d" % (folder_name, k)
        fh.write(",".join([peak_name, *["%.3g" % r for r in ratios]]) + "\n")

    fh.flush()

    return True, None

def collect_ratio_data(input_folder, result_folder):
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.SerialAnalyzer.DataUtils import serial_folder_walk
    global fh

    result_file = os.path.join(result_folder, "wyf-ratios.csv")

    fh = open(result_file, "w")

    serial_folder_walk(root_folder, get_wyf_ratios)

    fh.close()

if __name__ == '__main__':
    this_dir = os.path.dirname(os.path.abspath(__file__))
    lib_dir = os.path.dirname(this_dir)
    sys.path.append(lib_dir)

    import molass_legacy.KekLib, SerialAnalyzer, DataStructure, GuinierAnalyzer, Decomposer, Extrapolation
    from molass_legacy.SerialAnalyzer.DataUtils import get_pytools_folder

    root_folder = get_pytools_folder()
    input_folder = os.path.join(root_folder, "Data")
    result_folder = r"D:\TODO\20231218\wyf-ratios"
    collect_ratio_data(input_folder, result_folder)



# coding: utf-8
"""

    SingleComponent.py

    Copyright (c) 2018-2021, SAXS Team, KEK-PF

"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

PEAK_EVAL_RANGE_RATIO   = 0.05
DIFF_ALLOW_LIMIT_RATIO  = 0.02
USE_SMOOTHED_XRAY_CURVE = True
ACCEPTABLE_MIN_SCI  = 80
SCI_BOUNDARIES = [50, ACCEPTABLE_MIN_SCI]

if USE_SMOOTHED_XRAY_CURVE:
    from molass_legacy.Elution.CurveUtils import smoothed_curve_y

class SingleComponent:
    """
        created just after every mapping optimization
    """
    def __init__(self, mapper):
        self.mapper = mapper

    def compute_sci_ratio( self, debug=False ):
        mapper = self.mapper

        x = mapper.x_curve.x
        # y = mapper.x_curve.y
        y = mapper.x_curve_y_adjusted
        peak_top_y = y[mapper.x_curve.primary_peak_i]

        if USE_SMOOTHED_XRAY_CURVE:
            sy = smoothed_curve_y(mapper.x_curve, y=y)
            std_diff = np.std(y - sy)
            std_diff_ratio = std_diff/peak_top_y
        else:
            sy = y
            std_diff_ratio = 0

        allow_ratio = DIFF_ALLOW_LIMIT_RATIO + std_diff_ratio

        if debug:
            from matplotlib.patches import Rectangle
            from molass_legacy._MOLASS.SerialSettings import get_setting
            fig = plt.figure(figsize=(16,6))
            ax = fig.gca()
            ax_ = ax.twinx()
            in_folder = get_setting('in_folder')
            using = ' using smoothed xray elution' if USE_SMOOTHED_XRAY_CURVE else ''
            ax.set_title("Calulation of Single Component Indicator for " + in_folder + using, fontsize=16)

            sum_diff = 0
            set_label = True
            ax.plot( x, y, color='orange' )
            # ax.plot( x, mapper.x_curve.y, ':', color='orange' )
            if USE_SMOOTHED_XRAY_CURVE:
                ax.plot( x, sy, ':', color='red', label='smoothed xray' )
            ax.plot( x, mapper.mapped_vector, ':', color='blue', label='mapped uv' )
            xmin, xmax = x[[0, -1]]
            ax_.set_xlim(xmin, xmax)

            if USE_SMOOTHED_XRAY_CURVE:
                r = allow_ratio
                ax_.plot([xmin, xmax], [allow_ratio, allow_ratio], color='pink', label='adjusted allowance level')
                ax_.bar(x, abs(y-sy)/peak_top_y, color='gray', label='noise=abs(y - sy)/peak_top_y', alpha=0.2)
            r = DIFF_ALLOW_LIMIT_RATIO
            ax_.plot([xmin, xmax], [r, r], color='red', label='fixed allowance level')

            print('std_diff_ratio=', std_diff_ratio)
            print('mapper.peak_eval_ranges=', mapper.peak_eval_ranges)

        ratio_list = []
        for lower, mid, upper in mapper.peak_eval_ranges:
            slice_ = slice(lower, upper+1)
            y_  = sy[slice_]
            mv_ = mapper.mapped_vector[slice_]
            diff = np.abs( mv_ - y_ )

            ratio = diff / peak_top_y
            rwh = np.where( ratio > allow_ratio )[0]

            if debug:
                sum_diff += np.sum( diff )
                x_  = x[slice_]
                ax_.bar( x_, ratio, color='green', label='diff=(my - sy)/peak_top_y' if set_label else None, alpha=0.2)
                ax.plot( x_[rwh], mv_[rwh], 'o', color='yellow', label='badly matching points' if set_label else None )
                f, t = mapper.x_x[[lower, upper]]
                ymin, ymax = ax.get_ylim()
                p = Rectangle(
                        (f, ymin),  # (x,y)
                        t - f,   # width
                        ymax - ymin,    # height
                        facecolor   = 'cyan',
                        alpha       = 0.1,
                    )
                ax.add_patch(p)
                set_label = False

            if len(ratio) > 0:
                ratio = (len(y_) - len(rwh)) / len(ratio)
            else:
                ratio = np.nan
            ratio_list.append(ratio)

        if debug:
            print( 'sum_diff=', sum_diff )
            ax.legend(loc='upper left', fontsize=16)
            ax_.legend(loc='upper right', fontsize=16)
            plt.tight_layout()
            plt.show()

        # print( 'ratio_list=', ratio_list )
        assert len(ratio_list) == len(mapper.peak_eval_ranges)
        return ratio_list

    def compute_sci( self, debug=False ):
        sci_list = [ r*100 for r in self.compute_sci_ratio(debug=debug) ]
        return np.array(sci_list)



"""
    SecTheory.Felinger2005.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import basinhopping
from datetime import datetime
import molass_legacy.KekLib.DebugPlot as plt
from .SecCF import moving_zone_phi, two_site_phi
from .SecPDF import FftInvPdf, c

class MonoPoreModel:
    def __init__(self):
        self.pdf = c

    def fit(self, x, y, init_params=None):
        h_ = 100
        n1_ = len(x)//2
        t1_ = 10
        def obj_func(p):
            h, n1, t1 = p
            return np.sum((h*self.pdf(x, n1, t1) - y)**2)

        if init_params is None:
            init_params = [20, 60, 3]

        # ret = basinhopping(obj_func, (h_, n_, p1_, t1_, t2_))
        ret = basinhopping(obj_func, init_params)
        # ret = basinhopping(obj_func, init_params)
        return ret

    def __call__(self, x, h, n1, t1):
        return h*self.pdf(x, n1, t1)

class MonoPoreMovingZoneModel:
    def __init__(self):
        self.pdf = FftInvPdf(moving_zone_phi)

    def fit(self, x, y, init_params=None):
        def obj_func(p):
            h, n1, t1, N0, t0 = p
            # neg_penalty = min(0, n1)**2 + min(0, t1)**2 + min(0, N0)**2 + min(0, t0)**2
            neg_penalty = min(0, n1)**2 + min(0, N0)**2
            return np.sum((h*self.pdf(x, n1, t1, N0, t0) - y)**2) + neg_penalty*1e8

        if init_params is None:
            init_params = [20, 240, 1, 100, 10]

        # ret = basinhopping(obj_func, (h_, n_, p1_, t1_, t2_))
        ret = basinhopping(obj_func, init_params)
        # ret = basinhopping(obj_func, init_params)
        return ret

    def __call__(self, x, h, n1, t1, N0, t0):
        return h*self.pdf(x, n1, t1, N0, t0)

def demo():
    from .SecCF import simple_phi, moving_zone_phi
    from .SecPDF import c, FftInvPdf

    t = np.arange(300)

    npi = 10
    tpi = 10
    simple_pdf = FftInvPdf(simple_phi)

    moving_pdf = FftInvPdf(moving_zone_phi)
    N0 = 6
    t0 = 6

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        fig.suptitle("Proof of Numerical Inversion", fontsize=20)
        ax1.set_title("Simple", fontsize=16)
        ax1.plot(t, c(t, npi, tpi), label="formula")
        ax1.plot(t, simple_pdf(t, npi, tpi), label="numerical")
        ax1.legend()
        ax2.set_title("Moving Zone considered", fontsize=16)
        ax2.plot(t, moving_pdf(t, npi, tpi, N0, t0), label="numerical")
        ax2.legend()
        fig.tight_layout()
        plt.show()

def real_demo(in_folder, trimming=True, correction=True):
    from scipy.optimize import curve_fit
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Baseline.BaselineUtils import get_corrected_sd_impl
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from molass_legacy.Elution.CurveUtils import simple_plot
    from molass_legacy.Peaks.ElutionModels import egh
    from LPM import get_corrected
    from DataUtils import get_in_folder

    sp = StandardProcedure()
    sd = sp.load_old_way(in_folder)

    pre_recog = PreliminaryRecognition(sd)
    if trimming:
        sd_ = sd._get_analysis_copy_impl(pre_recog)
    else:
        sd_ = sd.get_copy()

    if correction:
        v2_copy = get_corrected_sd_impl(sd_, sd, pre_recog)
    else:
        v2_copy = sd_

    D, _, wv, ecurve = v2_copy.get_uv_data_separate_ly()

    ex = ecurve.x[10:-10]
    ey = get_corrected(ecurve.y[10:-10])

    p_init = ecurve.get_emg_peaks()[ecurve.primary_peak_no].get_params()
    popt1, pcov1 = curve_fit(egh, ex, ey, p_init)


    if False:
        model = MonoPoreModel()
        # init_params = [10, 150, 0.99, 0.1, 1]
        # init_params = [21.09085630844535, 57.05478375607099, 0.8134837337129833, 2.8055854804291713, 2.8055891557224237]
        # init_params = [21.09085438503959, 57.05479161605151, 0.4742816953836163, 2.8055865820081425, 2.805584983307841]
        # init_params = [21.0908540732968, 57.05479955016449, 0.8037212374299743, 2.8055851256770192, 2.805586333848221]
        # init_params = [21.0908535, 57.05480977, 2.80558488]
        init_params = [21.09085432546422, 57.05479510560458, 2.8055856104192967]
    elif in_folder.find("OA") > 0:
        model = MonoPoreMovingZoneModel()
        # init_params = [20, 240, 1, 100, 10]
        init_params = [21.279552638748612, 320.1301176968762, 1.8177109143998595e-07, 27.459933033640482, 157.87215347747625]
    else:
        assert False

    fh = open("params.csv", "w")

    ok_params = []
    min_value = None

    for k in range(10000):
        ret = model.fit(ex, ey, init_params=init_params)
        params = ret.x
        print("ret.fun=", ret.fun)

        if np.isnan(ret.fun):
            print("******************************** bad params =", params)
            init_params = ok_params[-1]
            continue

        if k % 1000 == 0:
            print(datetime.now().strftime("%Y-%m-%d %H:%M:%S"), [k], "---------------------------------------------")

        # if k % 100 == 0:
        if k == 0:
            with plt.Dp():
                in_folder = get_in_folder(in_folder)
                fig, ax = plt.subplots()
                ax.set_title("Two-site Stochastic Model Fitting for %s" % in_folder, fontsize=20)
                # simple_plot(ax, ecurve, legend=False)
                ax.plot(ex, ey, label="data")
                ax.plot(ex, egh(ex, *popt1), ":", lw=3, label="egh")
                ax.plot(ex, model(ex, *params), ":", lw=3, label="stochastic model")
                ax.legend(fontsize=16)
                fig.tight_layout()
                plt.show()

        if min_value is None or ret.fun < min_value:
            print([k], "------------------ ret.fun=", ret.fun)
            print("params=", params)
            fh.write("[%d] fv=%s\n" % (k, str(ret.fun)))
            fh.write("params=" + ",".join([str(p) for p in params]) + "\n")
            fh.flush()
            init_params = params
            min_value = ret.fun
            ok_params.append(params)

    fh.close()



"""
    Models/Stochastic/TriporeColumn.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
import molass_legacy.KekLib.DebugPlot as plt
from SecTheory.SecPDF import FftInvPdf
from SecTheory.SecCF import gec_tripore_phi
from molass_legacy.DataStructure.SvdDenoise import get_denoised_data
from molass_legacy.Models.ModelUtils import compute_cy_list, compute_area_props
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier
from molass_legacy.Models.Stochastic.Tripore import PORESIZES, POREPROPS

tripore_pdf = FftInvPdf(gec_tripore_phi)

def triporecolumn_func(me, mp, sv, pv, x, scale, N, T, Rg, x0):
    rho = Rg/sv
    rho[rho > 1] = 1
    np_ = N * pv * (1 - rho)**me
    tp_ = T*(1 - rho)**mp
    return scale*tripore_pdf(x, np_[0], tp_[0], np_[1], tp_[1], np_[2], tp_[2], x0)

class TriporeColumn:
    def __init__(self):
        self.N = 2000
        self.T = 8
        self.x0 = 50
        self.me = 1.5
        self.mp = 1.5
        self.sv = PORESIZES
        self.pv = POREPROPS

    def update_column_params(self, params):
        self.N = params[0]
        self.T = params[1]
        self.x0 = params[2]
        self.sv = params[3:6]
        self.pv = params[6:9]

    def estimate_psd(self, x, y, rgs, tRs, area_props, init_scales=None, use_basinhopping=False):
        N, T, x0 = [self.N, self.T, self.x0]
        def scale_objective(scales, debug=False):
            cy_list = []
            for rg, scale in zip(rgs, scales):
                cy = triporecolumn_func(self.me, self.mp, self.sv, self.pv, x, scale, N, T, rg, x0)
                cy_list.append(cy)
            ty = np.sum(cy_list, axis=0)
            props = compute_area_props(cy_list)
            if debug:
                with plt.Dp():
                    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 5))
                    fig.suptitle("objective debug")
                    ax1.plot(x, y)
                    for cy in cy_list:
                        ax1.plot(x, cy, ":")
                    ax1.plot(x, ty, ":", color="red")
                    ax2.bar(self.sv, self.pv, width=10)
                    fig.tight_layout()
                    plt.show()
            return np.sum((y - ty)**2) + np.sum((props - area_props)**2)

        if init_scales is None:
            init_scales = np.ones(len(rgs))
        scale_objective(init_scales, debug=True)
        res = minimize(scale_objective, init_scales, method='Nelder-Mead')
        scale_objective(res.x, debug=True)
        init_scales = res.x

        def objective(p, debug=False, title=None):
            N = p[0]
            T = p[1]
            x0 = p[2]
            sv = p[3:6]
            pv = p[6:9]
            scales = p[9:]
            cy_list = []
            for rg, scale in zip(rgs, scales):
                cy = triporecolumn_func(self.me, self.mp, sv, pv, x, scale, N, T, rg, x0)
                cy_list.append(cy)
            props = compute_area_props(cy_list)
            ty = np.sum(cy_list, axis=0)
            dsv = np.diff(sv)
            order_penalty = np.sum(dsv[dsv>0])
            if debug:
                print("N, T, x0=", N, T, x0)
                print("sv, pv=", sv, pv)
                print("dsv=", dsv, order_penalty)
                with plt.Dp():
                    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 5))
                    if title is None:
                        title = "objective debug"
                    fig.suptitle(title)
                    ax1.plot(x, y)
                    for cy, rg in zip(cy_list, rgs):
                        ax1.plot(x, cy, ":", label="Rg=%g" % rg)
                    ax1.plot(x, ty, ":", color="red", label="total")
                    ax1.legend()
                    ax2.bar(sv, pv, width=10)
                    fig.tight_layout()
                    plt.show()
            # return np.sum((y - ty)**2) + abs(1 - np.sum(pv)) + np.sum((props - area_props)**2) + max(100, np.std(sv))
            return np.sum((y - ty)**2) + abs(1 - np.sum(pv)) + np.sum((props - area_props)**2) + order_penalty

        init_params = np.concatenate([[self.N, self.T, self.x0], self.sv, self.pv, init_scales])
        bounds = [(0, 5000), (0, 10), (-100, 100)] + [(10, 300)]*3 + [(0, 1)]*3 + [(0, 100)]*len(rgs)
        objective(init_params, debug=True, title="before minimize")
        if use_basinhopping:
            minimizer_kwargs = dict(method='Nelder-Mead', bounds=bounds)
            res = basinhopping(objective, init_params, minimizer_kwargs=minimizer_kwargs, niter=10)
        else:
            method='Nelder-Mead' 
            res = minimize(objective, init_params, method=method, bounds=bounds)
        objective(res.x, debug=True, title="after minimize")
        return res.x

    def compute_cy_list(self, x, rgs, tripore_params):
        N, T, x0 = tripore_params[:3]
        sv = tripore_params[3:6]
        pv = tripore_params[6:9]
        scales = tripore_params[9:]
        cy_list = []
        for rg, scale in zip(rgs, scales):
            cy = triporecolumn_func(self.me, self.mp, sv, pv, x, scale, N, T, rg, x0)
            cy_list.append(cy)
        return cy_list

    def cover_missing_data(self, rg_list, last_rgs=None):
        if last_rgs is None:
            new_rg_list = rg_list
        else:
            new_rg_list = []
            for rg, last_rg in zip(rg_list, last_rgs):
                if rg is None:
                    new_rg_list.append(last_rg)
                else:
                    new_rg_list.append(rg)
        return np.array(new_rg_list)

    def get_curve(self, x, scale, *params):
        """
        all: N, T, me, mp, pore-distribution, Rg, scale, x0
        pore-distribution:
            pore sizes
            pore proportions

        mid: N, T, Rg, scale, x0  ---- implemented below

        min: Rg, scale, x0
        """
        return triporecolumn_func(self.me, self.mp, self.sv, self.pv, x, scale, *params)
    
def estimate_psd_impl(advanced_frame, devel=True):
    editor = advanced_frame.editor
    editor_frame = editor.get_current_frame()
    model = editor_frame.model
    print("estimate_psd_impl", model.get_name(), model.__class__)
    params_array = editor.get_current_params_array()

    fx = editor_frame.fx
    x = editor_frame.x
    y = editor_frame.y
    uv_y = editor_frame.uv_y

    D, E, qv, ecurve = editor.sd.get_xr_data_separate_ly()
    peak_region = ecurve.get_peak_region(sigma_scale=5)
    print("peak_region=", peak_region)

    def plot_cy_list(cy_list, title):
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title(title)
            ax.plot(x, y)
            for cy in cy_list:
                ax.plot(x, cy, ":", label="component-%d" % (len(cy_list)))
            ty = np.sum(cy_list, axis=0)
            ax.plot(x, ty, ":", color="red", label="model total")
            for x_ in peak_region:
                ax.axvline(x=x_+x[0], color="green")
            fig.tight_layout()
            plt.show()

    slice_ = slice(*[int(x_) for x_ in peak_region])
    num_components = len(params_array)
    print("num_components=", num_components)
    M_ = get_denoised_data(D[:,slice_], rank=num_components)
    E_ = E[:,slice_]
    cy_list = compute_cy_list(model, fx, params_array)
    plot_cy_list(cy_list, "initial state")

    min_msd = None
    column = TriporeColumn()
    init_scales = None
    last_rgs = None
    for k in range(3):
        C = np.array(cy_list)
        C_ = C[:,slice_]
        Cinv = np.linalg.pinv(C_)
        P_ = M_ @ Cinv
        Minv = np.linalg.pinv(M_)
        W = Minv @ P_
        Pe = np.sqrt(E_**2 @ W**2)

        rg_list = []
        for j, p_ in enumerate(P_.T):
            data = np.array([qv, p_, Pe[:,j]]).T
            sg = SimpleGuinier(data)
            rg_list.append(sg.Rg)

        rgs = column.cover_missing_data(rg_list, last_rgs=last_rgs)
        tRs = None
        area_props = compute_area_props(cy_list)
        tripore_params = column.estimate_psd(fx, y, rgs, tRs, area_props, init_scales=init_scales)
        temp_cy_list = column.compute_cy_list(fx, rgs, tripore_params)
        ty = np.sum(temp_cy_list, axis=0)
        msd = np.sum((y - ty)**2)
        print("msd=", msd)
        if min_msd is None or msd < min_msd:
            cy_list = temp_cy_list
            min_msd = msd
            column.update_column_params(tripore_params)
            plot_cy_list(cy_list, "%d-th estimate" % k)
            last_rgs = rgs
            init_scales = tripore_params[9:]
        else:
            break